/**************************************************************

CmpackCatFile.c (C-Munipack project)
Photometry file read/write routines
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <float.h>

#include "comfun.h"
#include "cmpack_common.h"
#include "cmpack_catfile.h"
#include "cathead.h"
#include "catstars.h"
#include "catparser.h"
#include "catselect.h"

/******************* Constants and definitions **********************/

#define BUFFSIZE 	8192

/*******************  Private data types   **************************/

/* Catalogue file context */
struct _CmpackCatFile 
{
	CmpackContext			gc;			/**< Global context */
	CmpackCatHeader			*head;		/**< File header */
	CmpackCatStarTab		*stars;		/**< Table of stars */
	CmpackCatSelectionTable	*select;	/**< Selection */
};

/********************** Local functions ****************************/

static void file_clean(void *ctx);

/* Alocates memory for photometry file and clears all fields. */
static CmpackCatFile *file_create(void)
{
	CmpackCatFile *f = gc_create(sizeof(CmpackCatFile));
	f->gc.clean = file_clean;
	f->head = cat_head_create();
	f->select = cat_seltab_create();
	f->stars = cat_stars_create();
	return f;
}

/* Frees all allocated memory in internal structures */
static void file_clean(void *ctx)
{
	CmpackCatFile *f = (CmpackCatFile*)ctx;

	cat_head_free(f->head);
	cat_seltab_free(f->select);
	cat_stars_free(f->stars);
	gc_clean(f);
}

/* Writes content of the file to the stream in XML format */
static void file_write(CmpackCatFile *f, FILE *to)
{
	/* Prolog */
	fprintf(to, "<?xml version=\"1.0\" encoding=\"ISO_8859-1\" ?>\n");
	fprintf(to, "<cat_file>\n");
	/* Write header */
	cat_head_write(f->head, to);
	/* Write selection */
	cat_seltab_write(f->select, to);
	/* Write stars */
	cat_stars_write(f->stars, to);
	/* Epilog */
	fprintf(to, "</cat_file>\n");
}

/*****************   Public functions    *******************/

/* Creates a catalogue file context */
CmpackCatFile *cmpack_cat_init(void)
{
	return file_create();
}

/* Load catalogue file */
int cmpack_cat_load(CmpackCatFile **pfile, const char *filename, CmpackLoadMode flags)
{
	int res;
	FILE *f;
	CmpackCatFile *file;

	*pfile = NULL;
	f = fopen(filename, "rb");
	if (!f) 
		return CMPACK_ERR_OPEN_ERROR;
	file = file_create();
	res = cat_loadxml(file, f, flags);
	if (res!=0) 
		cmpack_unref(file);
	else
		*pfile = file;
	fclose(f);
	return res;
}

/* Save catalogue file */
int cmpack_cat_save(CmpackCatFile *file, const char *filename)
{
	FILE *f;

	f = fopen(filename, "wb+");
	if (!f) 
		return CMPACK_ERR_OPEN_ERROR;

	file_write(file, f);
	fclose(f);
	return CMPACK_ERR_OK;
}

/* Save changes, close file and free allocated memory */
void cmpack_cat_clear(CmpackCatFile *f)
{
	cat_head_clear(f->head);
	cat_stars_clear(f->stars);
	cat_seltab_clear(f->select);
}

/* Makes copy of the file */
void cmpack_cat_copy(CmpackCatFile *fd, CmpackCatFile *fs)
{
	cat_head_copy(fd->head, fs->head);
	cat_stars_copy(fd->stars, fs->stars);
	cat_seltab_copy(fd->select, fs->select);
}

/* Make catalogue file from a photometry file */
int cmpack_cat_make(CmpackCatFile *fd, CmpackPhtFile *fs, int aperture)
{
	int i, w, h, id, mask, count;
	double x, y, jd, exptime;
	const char *filter;
	CmpackPhtMeasurement data;
	CmpackPhtStar *pstar;
	CmpackCatStar *cstar;

	cat_stars_clear(fd->stars);
	cat_seltab_clear(fd->select);

	jd = cmpack_pht_get_jd(fs);
	if (jd>0) 
		cmpack_cat_pkyd(fd, "jd", jd, JD_PRECISION, "");
	exptime = cmpack_pht_get_exptime(fs);
	if (exptime>0)
		cmpack_cat_pkyd(fd, "exptime", exptime, EXP_PRECISION, "");
	filter = cmpack_pht_get_filter(fs);
	if (filter)
		cmpack_cat_pkys(fd, "filter", filter, "");

	if (cmpack_pht_get_size(fs, &w, &h)) {
		cat_stars_set_size(fd->stars, w, h);
		count = cmpack_pht_nstar(fs);
		for (i=0; i<count; i++) {
			pstar = cmpack_pht_get_star(fs, i);
			if (pstar) {
				if (cmpack_pht_star_get_center(pstar, &x, &y)) {
					id = cmpack_pht_star_get_refid(pstar);
					if (id>=0) {
						cmpack_pht_star_get_measurement(pstar, aperture, &data);
						cstar = cat_star_new(id);
						mask = (data.mask & CMPACK_MASK_MAG) | CMPACK_MASK_POS;
						cat_star_set(cstar, mask, x, y, data.magnitude);
						cat_stars_add(fd->stars, cstar);
					}
				}
				cmpack_unref(pstar);
			}
		}
	}
	return 0;
}

/* Test if the file is a photometry file */
int cmpack_cat_test(const char *filename)
{
	int bytes, filesize;
	char buffer[2048];

	FILE *f = fopen(filename, "rb");
	if (f) {
		fseek(f, 0, SEEK_END);
		filesize = ftell(f);
		fseek(f, 0, SEEK_SET);
		bytes = (int)fread(buffer, 1, 2048, f);
		fclose(f);
		return cmpack_cat_test_buffer(buffer, bytes, filesize);
	}
	return 0;
}

/* Test if the file is a photometry file */
int cmpack_cat_test_buffer(const char *buffer, int buflen, int filesize)
{
	int res;
	char *aux;

	res = 0;
	if (buffer && buflen>16 && filesize>16) {
		/* Convert buffer to a null-terminated string */
		aux = (char*)cmpack_malloc((buflen+1)*sizeof(char));
		memcpy(aux, buffer, buflen);
		aux[buflen] = '\0';
		if (strstr(aux, "<?xml") && strstr(aux, "<cat_file")) {
			/* XML catalogue file */
			res = 1;
		}
		cmpack_free(aux);
	}
	return res;
}

/********************   Header manipulation    ***********************/

/* Get chart size */
int cmpack_cat_get_size(CmpackCatFile *f, int *width, int *height)
{
	return cat_stars_get_size(f->stars, width, height);
}

/* Get julian date of observation */
double cmpack_cat_get_jd(CmpackCatFile *file)
{
	if (cmpack_cat_gkys(file, "jd"))
		return cmpack_cat_gkyd(file, "jd", 0);
	else
		return 0;
}

/* Get exposure duration in seconds */
double cmpack_cat_get_exptime(CmpackCatFile *file)
{
	if (cmpack_cat_gkys(file, "exptime"))
		return cmpack_cat_gkyd(file, "exptime", 0);
	else
		return 0;
}

/* Get filter name */
const char *cmpack_cat_get_filter(CmpackCatFile *file)
{
	return cmpack_cat_gkys(file, "filter");
}

/* Get object's designation and coordinates */
const char *cmpack_cat_get_object(CmpackCatFile *file)
{
	return cmpack_cat_gkys(file, "object");
}
const char *cmpack_cat_get_objra(CmpackCatFile *file)
{
	return cmpack_cat_gkys(file, "ra2000");
}
const char *cmpack_cat_get_objdec(CmpackCatFile *file)
{
	return cmpack_cat_gkys(file, "dec2000");
}

/* Sets value and comment of specified key */
void cmpack_cat_pkys(CmpackCatFile *f, const char *key, const char *val, const char *com)
{
	cat_head_set(f->head, key, val, com);
}

/* Sets value and comment of specified key */
void cmpack_cat_pkyi(CmpackCatFile *f, const char *key, int val, const char *com)
{
	char buf[256];

	sprintf(buf, "%d", val);
	cat_head_set(f->head, key, buf, com);
}

/* Sets value and comment of specified key */
void cmpack_cat_pkyd(CmpackCatFile *f, const char *key, double val, int prec, const char *com)
{
	char buf[256];

	if (prec>16) 
		prec = 16;
	if (fabs(val)<DBL_EPSILON)
		strcpy(buf, "0");
	else if (fabs(val)<1e+9)
		sprintf(buf, "%.*f", prec, val);
	else
		sprintf(buf, "%.*e", prec, val);
	cat_head_set(f->head, key, buf, com);
}

/* Gets value and comment of specified key */
const char *cmpack_cat_gkys(CmpackCatFile *f, const char *key)
{
	CmpackCatHeadItem *item = cat_head_find(f->head, key);
	if (item)
		return cat_head_item_val(item);
	else
		return NULL;
}

/* Gets value and comment of specified key */
const char *cmpack_cat_gkyc(CmpackCatFile *f, const char *key)
{
	CmpackCatHeadItem *item = cat_head_find(f->head, key);
	if (item)
		return cat_head_item_com(item);
	else
		return NULL;
}

/* Gets value and comment of specified key */
int cmpack_cat_gkyi(CmpackCatFile *f, const char *key, int defval)
{
	CmpackCatHeadItem *item = cat_head_find(f->head, key);
	if (item) {
		const char *val = cat_head_item_val(item);
		if (val)
			return atoi(val);
	}
	return defval;
}

/* Gets value and comment of specified key */
double cmpack_cat_gkyd(CmpackCatFile *f, const char *key, double defval)
{
	CmpackCatHeadItem *item = cat_head_find(f->head, key);
	if (item) {
		const char *val = cat_head_item_val(item);
		if (val)
			return atof(val);
	}
	return defval;
}

/* Returns number of keys in active section */
int cmpack_cat_nkey(CmpackCatFile *f)
{
	return cat_head_count(f->head);
}

/* Get key, value and comment triplet by its index. The first record has index 0. */
int cmpack_cat_gkyn(CmpackCatFile *file, int index, const char **key, const char **val, const char **com)
{
	CmpackCatHeadItem *item = cat_head_get(file->head, index);
	if (item) {
		if (key) *key = cat_head_item_key(item);
		if (val) *val = cat_head_item_val(item);
		if (com) *com = cat_head_item_com(item);
		return 1;
	} else {
		if (key) *key = NULL;
		if (val) *val = NULL;
		if (com) *com = NULL;
		return 0;
	}
}

/* Deletes specified key */
void cmpack_cat_dkey(CmpackCatFile *f, const char *key)
{
	CmpackCatHeadItem *item = cat_head_find(f->head, key);
	if (item)
		cat_head_delete(f->head, item);
}

/********************   Table of stars   ***********************/

/* Returns number of stars defined */
int cmpack_cat_nstar(CmpackCatFile *f)
{
	return cat_stars_count(f->stars);
}

/* Find section by name */
CmpackCatStar *cmpack_cat_find_star(CmpackCatFile *f, int star_id)
{
	int index;

	if (star_id>=0) {
		index = cat_stars_find(f->stars, star_id);
		if (index>=0)
			return cmpack_addref(cat_stars_get(f->stars, index));
	}
	return NULL;
}

/* Append new star */
CmpackCatStar *cmpack_cat_add_star(CmpackCatFile *f, int star_id)
{
	if (star_id>=0) {
		CmpackCatStar *star = cat_star_new(star_id);
		if (star) {
			cat_stars_add(f->stars, star);
			return cmpack_addref(star);
		}
	}
	return NULL;
}

/* Get star by index */
CmpackCatStar *cmpack_cat_get_star(CmpackCatFile *f, int index)
{
	if (index>=0)
		return (CmpackCatStar*)cmpack_addref(cat_stars_get(f->stars, index));
	else
		return NULL;
}

/********************   Selection   ***********************/

/* Set selection type of all stars to CMPACK_SELECT_NONE */
void cmpack_cat_clear_selection(CmpackCatFile *f)
{
	cat_seltab_clear(f->select);
}

/* Get new reference to selection */
CmpackCatSelection *cmpack_cat_get_selection(CmpackCatFile *f, CmpackSelectionType type)
{
	return (CmpackCatSelection*)cmpack_addref(cat_seltab_get(f->select, type));
}

/* Set selection type by star identifier */
void cmpack_cat_set_star_type(CmpackCatFile *f, int star_id, CmpackSelectionType type)
{
	cat_seltab_set_star(f->select, star_id, type);
}

/* Get selection type by star identifier */
CmpackSelectionType cmpack_cat_get_star_type(CmpackCatFile *f, int star_id)
{
	return cat_seltab_get_star(f->select, star_id);
}
