/**************************************************************

phot_stars.c (C-Munipack project)
Table of stars
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "comfun.h"
#include "console.h"
#include "cmpack_common.h"
#include "cmpack_catfile.h"
#include "catstars.h"

#define ALLOC_BY 	64

/*********************   Datatypes   *******************/

/* Star context */
struct _CmpackCatStar
{
	CmpackContext gc;			/**< Global context */
	int id;						/**< Star identifier */
	int mask;					/**< Mask of valid fields */
	double x, y;				/**< Position */
	double mag;					/**< Magnitude */
};

/* Table of stars */
struct _CmpackCatStarTab
{
	int nx, ny;					/**< Chart size */
	int count;					/**< Number of stars */
	int capacity;				/**< Number of allocated records */
	CmpackCatStar **list;		/**< List of stars */
};

/********************   Local functions    ***********************/

/* Create a new star */
static CmpackCatStar *star_create(int id)
{
	CmpackCatStar *st = (CmpackCatStar*)gc_create(sizeof(CmpackCatStar));
	st->id = id;
	return st;
}

/* Copy star attributes */
static void star_copy(CmpackCatStar *to, const CmpackCatStar *from)
{
	to->id = from->id;
	to->mag = from->mag;
	to->mask = from->mask;
	to->x = from->x;
	to->y = from->y;
}

/* Dump star attributes */
static void star_dump(const CmpackCatStar *star, CmpackConsole *con)
{
	char aux[MAXLINE], aux1[MAXLINE];

	sprintf(aux, "\t\t");
	sprintf(aux1, "Id=%d", star->id);
	strcat(aux, aux1);
	if ((star->mask & CMPACK_MASK_POS)!=0) {
		sprintf(aux1, "; X=%.*f; Y=%.*f", POS_PRECISION, star->x, POS_PRECISION, star->y);
		strcat(aux, aux1);
	}
	if ((star->mask & CMPACK_MASK_MAG)!=0) {
		sprintf(aux1, "; Mag=%.*f", MAG_PRECISION, star->mag);
		strcat(aux, aux1);
	}
	printout(con, 1, aux);
}

/* Save star to file */
static void star_write(const CmpackCatStar *star, FILE *to)
{	
	fprintf(to, "<s id=\"%d\"", star->id);
	if ((star->mask & CMPACK_MASK_POS)!=0)
		fprintf(to, " x=\"%.*f\" y=\"%.*f\"", POS_PRECISION, star->x, POS_PRECISION, star->y);
	if ((star->mask & CMPACK_MASK_MAG)!=0)
		fprintf(to, " m=\"%.*f\"", MAG_PRECISION, star->mag);
	fprintf(to, "/>\n");
}

/********************   Public functions    ***********************/

/* Initialize the table of star */
CmpackCatStarTab *cat_stars_create(void)
{
	return (CmpackCatStarTab*)cmpack_calloc(1, sizeof(CmpackCatStarTab));
}

/* Free allocated memory */
void cat_stars_free(CmpackCatStarTab *tab)
{
	int i;

	for (i=0; i<tab->count; i++)
		cmpack_unref(tab->list[i]);
	cmpack_free(tab->list);
	cmpack_free(tab);
}

/* Dump table of stars */
void cat_stars_dump(CmpackCatStarTab *tab, CmpackConsole *con)
{
	int i;
	char aux[MAXLINE];
	
	sprintf(aux, "\tStar: %d items\n", tab->count);
	printout(con, 1, aux);
	for (i=0; i<tab->count; i++) 
		star_dump(tab->list[i], con);
}

/* Dump table of stars */
int cat_stars_find(CmpackCatStarTab *tab, int star_id)
{
	int i;
	
	for (i=0; i<tab->count; i++) {
		if (tab->list[i]->id == star_id)
			return i;
	}
	return 0;
}

/* Write all stars to file */
void cat_stars_write(CmpackCatStarTab *tab, FILE *to)
{
	int i;
	
	fprintf(to, "<stars width=\"%d\" height=\"%d\">\n", tab->nx, tab->ny);
	for (i=0; i<tab->count; i++)
		star_write(tab->list[i], to);
	fprintf(to, "</stars>\n");
}

/* Make copy of a table of stars */
void cat_stars_copy(CmpackCatStarTab *dsttab, const CmpackCatStarTab *srctab)
{
	int i;
	
	cat_stars_clear(dsttab);
	dsttab->nx = srctab->nx;
	dsttab->ny = srctab->ny;
	if (srctab->list && srctab->count>0) {
		dsttab->count = dsttab->capacity = srctab->count;
		dsttab->list = (CmpackCatStar**)cmpack_realloc(dsttab->list, dsttab->capacity*sizeof(CmpackCatStar*));
		for (i=0; i<dsttab->count; i++) {
			dsttab->list[i] = star_create(srctab->list[i]->id);
			star_copy(dsttab->list[i], srctab->list[i]);
		}
	}
}

/* Clear table of stars */
void cat_stars_clear(CmpackCatStarTab *tab)
{
	int i;

	tab->nx = tab->ny = 0;
	for (i=0; i<tab->count; i++)
		cmpack_unref(tab->list[i]);
	cmpack_free(tab->list);
	tab->list = NULL;
	tab->count = tab->capacity = 0;
}

/* Add new star */
int cat_stars_add(CmpackCatStarTab *tab, CmpackCatStar *star)
{
	int i = tab->count;
	if (tab->count>=tab->capacity) {
		tab->capacity += ALLOC_BY;
		tab->list = (CmpackCatStar**)cmpack_realloc(tab->list, tab->capacity*sizeof(CmpackCatStar*));
	}
	tab->list[i] = star;
	tab->count++;
	return i;	
}

/************************ Public interface ****************************/ 

/* Creates a new star and returns its index */
CmpackCatStar *cat_star_new(int id)
{
	return star_create(id);
}

/* Get star indentifier */
int cmpack_cat_star_get_id(CmpackCatStar *s)
{
	return s->id;
}

/* Set position of a star */
void cmpack_cat_star_set_center(CmpackCatStar *s, double x, double y)
{
	s->x = x;
	s->y = y;
	s->mask |= CMPACK_MASK_POS;
}

/* Get position of a star */
int cmpack_cat_star_get_center(CmpackCatStar *s, double *x, double *y)
{
	if (s->mask & CMPACK_MASK_POS) {
		if (x) *x = s->x;
		if (y) *y = s->y;
		return 1;
	} else {
		if (x) *x = 0;
		if (y) *y = 0;
		return 0;
	}
}

/* Set magnitude of a star */
void cmpack_cat_star_set_magnitude(CmpackCatStar *s, double mag)
{
	s->mag = mag;
	s->mask |= CMPACK_MASK_MAG;
}

/* Get magnitude of a star */
int cmpack_cat_star_get_magnitude(CmpackCatStar *s, double *mag)
{
	if (s->mask & CMPACK_MASK_MAG) {
		if (mag) *mag = s->mag;
		return 1;
	} else {
		if (mag) *mag = 0;
		return 0;
	}
}

/* Set chart size */
void cat_stars_set_size(CmpackCatStarTab *tab, int width, int height)
{
	tab->nx = width;
	tab->ny = height;
}

/* Get chart size */
int cat_stars_get_size(CmpackCatStarTab *tab, int *width, int *height)
{
	if (width) *width = tab->nx;
	if (height) *height = tab->ny;
	return 1;
}

/* Get number of stars */
int cat_stars_count(CmpackCatStarTab *tab)
{
	return tab->count; 
}

/* Get star context by its index */
CmpackCatStar *cat_stars_get(CmpackCatStarTab *tab, int index)
{
	if (index>=0 && index<tab->count)
		return tab->list[index];
	else
		return NULL;
}

/* Creates a new star and returns its index */
void cat_star_set(CmpackCatStar *star, int mask, double x, double y, double mag)
{
	if (mask & CMPACK_MASK_POS) {
		star->mask |= CMPACK_MASK_POS;
		star->x = x;
		star->y = y;
	}
	if (mask & CMPACK_MASK_MAG) {
		star->mask |= CMPACK_MASK_MAG;
		star->mag = mag;
	}
}
