/**************************************************************

dark.c (C-Munipack project)
Dark correction
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "comfun.h"
#include "console.h"
#include "ccdfile.h"
#include "cmpack_common.h"
#include "cmpack_dark.h"

/*******************   DATA TYPES   ***************************/

/* Dark correction context */
struct _CmpackDarkCorr
{
	CmpackContext gc;			/**< Global context */
	CmpackConsole *con;			/**< Console context */
	CmpackBorder border;		/**< Border size */
	int scaling;				/**< Use dark-frame scaling */
	int nx;						/**< Width of dark-frame */
	int ny;						/**< Height of dark-frame */
	double *ddata;				/**< Data of dark-frame */
	int scalable;				/**< Is dark frame scalable? */
	double exptime;				/**< Exposure duration for dark-frame */
};

/********************   LOCAL FUNCTIONS   *****************************/

/* Frees allocated memory */
static void dark_clean(void *ctx)
{
	CmpackDarkCorr *lc = (CmpackDarkCorr*)ctx;

	cmpack_free(lc->ddata);
	cmpack_unref(lc->con);
	gc_clean(ctx);
}

/* Computes the output image from source image and dark image (out=sci-dark) */
static int dark_dark(CmpackDarkCorr *lc, CmpackCcdFile *infile, CmpackCcdFile *outfile)
{
	int res, i, nx, ny, x, y, left, top, right, bottom, underflow, overflow;
	double exptime, scale, *sdata, minvalue, maxvalue;
	char msg[MAXLINE];
	CmpackBitpix bitpix;

	/* Check image size */
	nx = ny = 0;
	cmpack_ccd_get_size(infile, &nx, &ny);
	if (nx<=0 || nx>=65536 || ny<=0 || ny>=65536) 
		return CMPACK_ERR_INVALID_SIZE;
	if (lc->nx!=nx || lc->ny!=ny) 
		return CMPACK_ERR_DIFF_SIZE_DARK;

	/* Get data format */
	bitpix = cmpack_ccd_get_bitpix(infile);
	cmpack_ccd_get_range(infile, &minvalue, &maxvalue);
	if (bitpix==CMPACK_BITPIX_UNKNOWN) {
		printout(lc->con, 0, "Unsupported image data format.");
		return CMPACK_ERR_READ_ERROR;
	}
	if (is_debug(lc->con)) {
		sprintf(msg, "Data format  : %s", pixformat(bitpix));
		printout(lc->con, 1, msg);
		sprintf(msg, "Min. value   : %.12g", minvalue); 
		printout(lc->con, 1, msg);
		sprintf(msg, "Max. value   : %.12g", maxvalue); 
		printout(lc->con, 1, msg);
		sprintf(msg, "Border     : %d %d %d %d", lc->border.left, lc->border.top,
			lc->border.right, lc->border.bottom);
		printout(lc->con, 1, msg);
	}

	/* Read and check exposure duration */
	exptime = 0.0;
	if (lc->scaling && lc->scalable) {
		if (cmpack_ccd_get_exptime(infile, &exptime)!=0) {
			printout(lc->con, 1, "Cannot read exposure duration.");
			return CMPACK_ERR_READ_ERROR;
		}
		if (exptime<=0.0) {
			printout(lc->con, 1, "Invalid exposure duration.");
			return CMPACK_ERR_INVALID_EXPTIME;
		}
		sprintf(msg,"Exposure duration: %.2f s", exptime);
		printout(lc->con, 0, msg);
	}

	/* Read exposure data */
	sdata = cmpack_malloc(nx*ny*sizeof(double));
	res = cmpack_ccd_get_image(infile, CMPACK_BITPIX_DOUBLE, 0, 0, sdata);
	if (res!=0) {
		cmpack_free(sdata);
		return res;
	}

	/* Dark subtraction and range checking */
	underflow = overflow = 0;
	left = lc->border.left;
	top = lc->border.top;
	right = nx - lc->border.right;
	bottom = ny - lc->border.bottom;
	scale = (lc->scaling && lc->scalable ? exptime/lc->exptime : 1.0);
	for (y=0;y<ny;y++) {
		for (x=0;x<nx;x++) {
			i = x + y*nx;
			if (x>=left && x<right && y>=top && y<bottom) {
				sdata[i] -= scale*lc->ddata[i];
				if (sdata[i]<minvalue) {
					sdata[i] = minvalue;
					underflow = 1;
				}
				if (sdata[i]>maxvalue) {
					sdata[i] = maxvalue;
					overflow = 1;
				}
			} else {
				sdata[i] = 0.0;
			}
		}
	}

	/* Store image data */
	res = ccd_write_image(outfile, CMPACK_BITPIX_DOUBLE, sdata);
	cmpack_free(sdata);

	/* Update frame header */
	ccd_write_history(outfile, "Dark frame subtracted.");

	/* Normal return */
	if (overflow)
		printout(lc->con, 1, "An overflow has been occurred during computation.");
	if (underflow)
		printout(lc->con, 1, "An underflow has been occurred during computation.");
	return res;
}

/**********************    PUBLIC FUNCTIONS   ***********************************/

/* Initializes the variables and loads the dark-frame */
CmpackDarkCorr *cmpack_dark_init(void)
{
	CmpackDarkCorr *lc = gc_create(sizeof(CmpackDarkCorr));
	lc->gc.clean = &dark_clean;
	return lc;
}

/* Attach console */
void cmpack_dark_set_console(CmpackDarkCorr *ctx, CmpackConsole *con)
{
	if (con!=ctx->con) {
		cmpack_unref(ctx->con);
		ctx->con = cmpack_addref(con);
	}
}

/* Enable or disable dark-frame scaling */
void cmpack_dark_set_scaling(CmpackDarkCorr *ctx, int scaling)
{
	ctx->scaling = scaling!=0;
}

/* Get status of dark-frame scaling */
int cmpack_dark_get_scaling(CmpackDarkCorr *ctx)
{
	return ctx->scaling;
}

/* Set image border */
void cmpack_dark_set_border(CmpackDarkCorr *lc, const CmpackBorder *border)
{
	if (border)
		lc->border = *border;
	else
		memset(&lc->border, 0, sizeof(CmpackBorder));
}

/* Set image border */
void cmpack_dark_get_border(CmpackDarkCorr *lc, CmpackBorder *border)
{
	if (border)
		*border = lc->border;
}

/* Initializes the variables and loads the dark-frame */
int cmpack_dark_rdark(CmpackDarkCorr *lc, CmpackCcdFile *dark)
{
	int res, nx, ny, scalable;
	double exptime;

	/* Clear previous dark frame */
	cmpack_free(lc->ddata);
	lc->ddata = NULL;
	lc->nx = lc->ny = lc->scalable = 0;
	lc->exptime = 0.0;

    /* Check file name */
    if (!dark) {
        printout(lc->con, 0, "Invalid dark frame context.");
 	    return CMPACK_ERR_INVALID_PAR;
    }

    /* Read and check dimensions */
	nx = ny = 0;
	cmpack_ccd_get_size(dark, &nx, &ny);
	if (nx<=0 || nx>=65536 || ny<=0 || ny>=65536) {
		printout(lc->con, 1, "Invalid dimensions of the dark frame.");
		return CMPACK_ERR_INVALID_SIZE;
	}
	lc->nx = nx;
	lc->ny = ny;

	/* Is the dark frame scalable? */
	exptime = 0.0;
	if (cmpack_ccd_get_exptime(dark, &exptime)!=0) {
		printout(lc->con, 1, "Cannot read exposure duration in the dark frame.");
		return CMPACK_ERR_READ_ERROR;
	}
	if (exptime<=0.0) {
		printout(lc->con, 1, "Invalid exposure duration in the dark frame.");
		return CMPACK_ERR_INVALID_EXPTIME;
	}
	if (cmpack_ccd_get_bool(dark, "SCALABLE", &scalable)==0) {
		lc->scalable = scalable && exptime>0;
		lc->exptime = exptime;
	}

	/* Read dark frame data */
	lc->ddata = cmpack_malloc(lc->nx*lc->ny*sizeof(double));
	res = cmpack_ccd_get_image(dark, CMPACK_BITPIX_DOUBLE, 0, 0, lc->ddata);
	if (res!=0) {
		cmpack_free(lc->ddata);
		lc->ddata = NULL;
		return res;
	}

	/* Print parameters */
	if (is_debug(lc->con)) {
		printout(lc->con, 1, "Dark correction frame:");
		printpari(lc->con, "Width", 1, lc->nx);
		printpari(lc->con, "Height", 1, lc->ny);
		printpard(lc->con, "Exp. time", 1, lc->exptime, 2);
		if (lc->scaling) {
			if (lc->scalable && lc->exptime>0) 
				printout(lc->con, 1, "This is scalable dark, using advanced calibration.");
			else
				printout(lc->con, 1, "Dark frame is not scalable, using standard calibration!");
		}
	}
    return 0;
}

/* Computes the output image from source image and dark image (out=sci-dark) */
int cmpack_dark(CmpackDarkCorr *lc, CmpackCcdFile *file)
{
	/* Check dark file */
	if (!lc->ddata) {
		printout(lc->con, 0, "Missing dark frame.");
		return CMPACK_ERR_NO_DARK_FRAME;
	}

  	/* Check parameters */
	if (!file) {
		printout(lc->con, 0, "Invalid frame context.");
		return CMPACK_ERR_INVALID_PAR;
	}

	/* Dark correction */
	return dark_dark(lc, file, file);
}


/* Computes the output image from source image and dark image (out=sci-dark) */
int cmpack_dark_ex(CmpackDarkCorr *lc, CmpackCcdFile *infile, CmpackCcdFile *outfile)
{
	int res, nx, ny;
	CmpackBitpix bitpix;

	/* Check dark file */
	if (!lc->ddata) {
		printout(lc->con, 0, "Missing dark frame.");
		return CMPACK_ERR_NO_DARK_FRAME;
	}

  	/* Check parameters */
	if (!infile) {
		printout(lc->con, 0, "Invalid input frame context.");
		return CMPACK_ERR_INVALID_PAR;
	}
	if (!outfile) {
		printout(lc->con, 0, "Invalid output frame context.");
		return CMPACK_ERR_INVALID_PAR;
	}

	/* Check image size */
	nx = ny = 0;
	cmpack_ccd_get_size(infile, &nx, &ny);
	if (nx<=0 || nx>=65536 || ny<=0 || ny>=65536) {
		printout(lc->con, 0, "Invalid frame size.");
		return CMPACK_ERR_INVALID_SIZE;
	}

	/* Get data format */
	bitpix = cmpack_ccd_get_bitpix(infile);
	if (bitpix==CMPACK_BITPIX_UNKNOWN) {
		printout(lc->con, 0, "Unsupported image data format.");
		return CMPACK_ERR_READ_ERROR;
	}

	/* Init output image */
	res = ccd_prepare(outfile, nx, ny, bitpix);
	if (res!=0) 
		return res;

	/* Copy file header */
	res = ccd_copy_header(outfile, infile);
	if (res!=0) 
		return res;

	/* Dark correction */
	return dark_dark(lc, infile, outfile);
}
