/**************************************************************

hcor.c (C-Munipack project)
Heliocentric correction
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>

#include "comfun.h"
#include "console.h"
#include "trajd.h"
#include "cmpack_common.h"
#include "cmpack_hcorr.h"

/*******************   DATA TYPES   ***************************/

/* Heliocentric correction context */
struct _CmpackHelCorr
{
	CmpackContext gc;			/**< Global context */
	CmpackConsole *con;			/**< Console */
	int mask;					/**< Which fields are valid? */
	double ra;					/**< Object's right ascension */
	double dec;					/**< Object's declination */
};

/**********************   PRIVATE FUNCTIONS   ***************************/

static void hcorr_clean(void *ctx)
{
	CmpackHelCorr *lc = (CmpackHelCorr*) ctx;

	cmpack_unref(lc->con);
	gc_clean(ctx);
}

/**********************   PUBLIC FUNCTIONS   ********************************/

/* Context initialization */
CmpackHelCorr *cmpack_hcorr_init(void)
{
	CmpackHelCorr *lc = gc_create(sizeof(CmpackHelCorr));
	lc->gc.clean = &hcorr_clean;
	return lc;
}

/* Attach console */
void cmpack_hcorr_set_console(CmpackHelCorr *ctx, CmpackConsole *con)
{
	if (con!=ctx->con) {
		cmpack_unref(ctx->con);
		ctx->con = cmpack_addref(con);
	}
}

/* Set object's right ascension */
void cmpack_hcorr_set_ra(CmpackHelCorr *ctx, double ra)
{
	ctx->mask |= CMPACK_MASK_RA;
	ctx->ra = ra;
}

/* Get object's right ascension */
int cmpack_hcorr_get_ra(CmpackHelCorr *ctx, double *pra)
{
	if (ctx->mask & CMPACK_MASK_RA) {
		if (pra) *pra = ctx->ra;
		return 0;
	} else {
		if (pra) *pra = 0.0;
		return CMPACK_ERR_UNDEF_VALUE;
	}
}

/* Set object's declination */
void cmpack_hcorr_set_dec(CmpackHelCorr *ctx, double dec)
{
	ctx->mask |= CMPACK_MASK_DEC;
	ctx->dec = dec;
}

/* Get object's declination */
int cmpack_hcorr_get_dec(CmpackHelCorr *ctx, double *pdec)
{
	if (ctx->mask & CMPACK_MASK_DEC) {
		if (pdec) *pdec = ctx->dec;
		return 0;
	} else {
		if (pdec) *pdec = 0.0;
		return CMPACK_ERR_UNDEF_VALUE;
	}
}

int cmpack_hcorr(CmpackHelCorr *lc, CmpackTable *tab, int jd_column, CmpackHCorrFlags flags)
{
	int    i, add_value, no_modify, reverse, count, hc_column;
	char   msg[MAXLINE];
	double ls, rs, la, be, jdin, jdout, hcor;

	add_value = (flags & CMPACK_HCORR_ADDVALUE)!=0;
	no_modify = (flags & CMPACK_HCORR_NOMODIFY)!=0;
	reverse   = (flags & CMPACK_HCORR_REVERSE)!=0;

	if (jd_column<0 || jd_column>=cmpack_tab_ncolumns(tab))
		return CMPACK_ERR_INVALID_PAR;

	/* Check parameters */
	if ((lc->mask & CMPACK_MASK_RA)==0) {
		printout(lc->con, 0, "Missing object's right ascension.");
		return CMPACK_ERR_NO_OBJ_COORDS;
	}
	if ((lc->mask & CMPACK_MASK_DEC)==0) {
		printout(lc->con, 0, "Missing object's declination.");
		return CMPACK_ERR_NO_OBJ_COORDS;
	}
	/* Print parameters */
	if (is_debug(lc->con)) {
		printout(lc->con, 1, "Configuration parameters:");
	  	sprintf(msg, "\tR.A. = %.7f", lc->ra);
	  	printout(lc->con, 1, msg);
		sprintf(msg, "\tDEC. = %.7f", lc->dec);
	  	printout(lc->con, 1, msg);
  	}
	/* Compute object's ecliptic coordinates */
	cmpack_rdtolb(lc->ra, lc->dec, &la, &be);

	/* Column which contains the correction value */
	hc_column = -1;
	if (add_value) {
		hc_column = cmpack_tab_find_column(tab, "HELCOR");
		if (hc_column<0)
			hc_column = cmpack_tab_add_column(tab, "HELCOR", CMPACK_COLUMN_DOUBLE, JD_PRECISION);
	}

	/* Rename JD column to JDHEL or JDGEO */
	if (!no_modify) {
		CmpackTabColumn *col = cmpack_tab_get_column(tab, jd_column);
		cmpack_tab_column_set_name(col, (!reverse ? "JDHEL" : "JDGEO"));
		cmpack_unref(col);
	}

	count = cmpack_tab_nrows(tab);
	for (i=0; i<count; i++) {
		cmpack_tab_gtdd(tab, i, jd_column, &jdin);
		if (jdin<=0)
			return CMPACK_ERR_INVALID_JULDAT;
	
		cmpack_sun(jdin, &ls, &rs);
		hcor = cmpack_hcor2(la, be, ls, rs);
		if (no_modify) {
			jdout = jdin;
			if (is_debug(lc->con)) {
				sprintf(msg, "%.7f -> %.7f", jdin, hcor);
				printout(lc->con, 1, msg);
			}
		} else {		
			if (reverse) {
				jdout = jdin - hcor;
				if (is_debug(lc->con)) {
					sprintf(msg, "%.7f - (%.7f) = %.7f", jdin, hcor, jdout);
					printout(lc->con, 1, msg);
				}
			} else {
				jdout = jdin + hcor;
				if (is_debug(lc->con)) {
					sprintf(msg, "%.7f + (%.7f) = %.7f", jdin, hcor, jdout);
					printout(lc->con, 1, msg);
				}
			}
			cmpack_tab_ptdd(tab, i, jd_column, jdout);
		}
		if (hc_column>0) 
			cmpack_tab_ptdd(tab, i, hc_column, hcor);
	}  
 	return 0;
}


int cmpack_hcorr_jd(CmpackHelCorr *lc, double jd, double *phcor)
{
	char   msg[MAXLINE], buf[MAXLINE];
	double ls, rs, la, be;

	/* Check parameters */
	if ((lc->mask & CMPACK_MASK_RA)==0) {
		printout(lc->con, 0, "Missing object's right ascension.");
		return CMPACK_ERR_NO_OBJ_COORDS;
	}
	if ((lc->mask & CMPACK_MASK_DEC)==0) {
		printout(lc->con, 0, "Missing object's declination.");
		return CMPACK_ERR_NO_OBJ_COORDS;
	}
	/* Print parameters */
	if (is_debug(lc->con)) {
		printout(lc->con, 1, "Configuration parameters:");
	  	cmpack_ratostr(lc->ra, buf, MAXLINE);
	  	sprintf(msg, "\tR.A. = %s", buf);
	  	printout(lc->con, 1, msg);
	  	cmpack_dectostr(lc->dec, buf, MAXLINE);
	  	sprintf(msg, "\tDEC. = %s", buf);
	  	printout(lc->con, 1, msg);
  	}
	
	if (phcor) {
		cmpack_rdtolb(lc->ra, lc->dec, &la, &be);
		cmpack_sun(jd, &ls, &rs);
		*phcor = cmpack_hcor2(la, be, ls, rs);
	}  

	/* Normal return */
 	return 0;
}
