/**************************************************************

kombine.c (C-Munipack project)
Combine frames
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "config.h"
#include "comfun.h"
#include "console.h"
#include "trajd.h"
#include "ccdfile.h"
#include "cmpack_common.h"
#include "cmpack_kombine.h"

/*******************   DATA TYPES   ***************************/

/* Kombine output file context */
struct _CmpackKombine
{
	CmpackContext gc;			/**< Global context */
	CmpackConsole *con;			/**< Console */
	CmpackCcdFile *outfile;		/**< Output file context */
	CmpackBitpix bitpix;   		/**< Required output image data type */
	CmpackBorder border;		/**< Border size */
	CmpackBitpix in_bitpix;   	/**< Output data format */
	int in_width;             	/**< Width of output frame */
	int in_height;             	/**< Height of output frame */
	double sjd;					/**< Accumulated JD */
	double exptime;				/**< Accumulated exposure duration */
	int nframe;					/**< Number of source frames */
	double *data;				/**< Accumulation buffer */
	char *stat;					/**< Status bits for each pixel */
	CmpackCcdFile *header;		/**< Header of first frame */
};

/*******************   LOCAL FUNCTIONS   ***************************/

#define I(x,y)	ccd[x+nx*y]
#define D(x,y)	lc->data[x+nx*y]
#define S(x,y)	lc->stat[x+nx*y]

/* Clean frame context */
static void kombine_clean(void *ctx)
{
	CmpackKombine *lc = (CmpackKombine*)ctx;
	
	cmpack_unref(lc->header);
	cmpack_free(lc->data);
	cmpack_free(lc->stat);
	cmpack_unref(lc->outfile);
	cmpack_unref(lc->con);
	gc_clean(ctx);
}

/*****************   PUBLIC FUNCTIONS   ******************************/

/* Reads first image and opens destination file */
CmpackKombine *cmpack_kombine_init(void)
{
	CmpackKombine *lc = gc_create(sizeof(CmpackKombine));
	lc->gc.clean = &kombine_clean;
	lc->bitpix = CMPACK_BITPIX_AUTO;
	lc->header = cmpack_ccd_new();
	return lc;
}

/* Attach console */
void cmpack_kombine_set_console(CmpackKombine *ctx, CmpackConsole *con)
{
	if (con!=ctx->con) {
		cmpack_unref(ctx->con);
		ctx->con = cmpack_addref(con);
	}
}

/* Set output image data format */
void cmpack_kombine_set_bitpix(CmpackKombine *lc, CmpackBitpix bitpix)
{
	lc->bitpix = bitpix;
}

/* Set output image data format */
CmpackBitpix cmpack_kombine_get_bitpix(CmpackKombine *lc)
{
	return lc->bitpix;
}

/* Set image border */
void cmpack_kombine_set_border(CmpackKombine *lc, const CmpackBorder *border)
{
	if (border)
		lc->border = *border;
	else
		memset(&lc->border, 0, sizeof(CmpackBorder));
}

/* Set image border */
void cmpack_kombine_get_border(CmpackKombine *lc, CmpackBorder *border)
{
	if (border)
		*border = lc->border;
}

/* Reads first image and opens destination file */
int cmpack_kombine_open(CmpackKombine *lc, CmpackCcdFile *outfile)
{
	char msg[256];

	/* Print configuration parameters */
	if (is_debug(lc->con)) {
		printout(lc->con, 1, "Kombine parametets:");
		printpari(lc->con, "BitPix", 1, lc->bitpix);
		sprintf(msg, "Border     : %d %d %d %d", lc->border.left, lc->border.top,
			lc->border.right, lc->border.bottom);
		printout(lc->con, 1, msg);
	}
	/* Store file name */
	lc->outfile = cmpack_addref(outfile);
	ccd_clear_header(lc->header);
    /* Normal return */
    return 0;
}

/* Reads first image and opens destination file */
int cmpack_kombine_read(CmpackKombine *lc, CmpackCcdFile *ccdfile, CmpackPhtFile *phtfile)
{
	int		res, nx, ny, i, j, ii, jj, offsetx, offsety;
    double	*ccd, jd, exptime, x, dx, dy, minvalue, maxvalue;
    char	msg[MAXLINE];
	CmpackDateTime dt;
	CmpackBitpix bitpix;

    /* Check source files */
    if (!ccdfile) {
        printout(lc->con, 0, "Invalid image frame context.");
 	    return CMPACK_ERR_INVALID_PAR;
    }
    if (!phtfile) {
        printout(lc->con, 0, "Invalid photometry file context.");
 	    return CMPACK_ERR_INVALID_PAR;
    }

	/* Read frame offset */
	if (!cmpack_pht_get_offset(phtfile, &dx, &dy)) {
	    sprintf(msg, "Missing frame offset information in the photometry file.");
	    printout(lc->con, 0, msg);
        return CMPACK_ERR_READ_ERROR;
	}
	offsetx = (int)(dx>=0.0 ? dx+0.5 : dx-0.5);
	offsety = (int)(dy>=0.0 ? dy+0.5 : dy-0.5);

    /* Read parameters */
	nx = ny = 0;
    cmpack_ccd_get_size(ccdfile, &nx, &ny);
    if (nx==0 || ny==0) 
	    return CMPACK_ERR_INVALID_SIZE;

	bitpix = cmpack_ccd_get_bitpix(ccdfile);
	cmpack_ccd_get_range(ccdfile, &minvalue, &maxvalue);
	if (bitpix==0)
		return CMPACK_ERR_INVALID_BITPIX;

	if (cmpack_ccd_get_datetime(ccdfile, &dt)!=0)
		return CMPACK_ERR_READ_ERROR;
	jd = cmpack_encodejd(&dt);
	if (jd<=0.0)
		return CMPACK_ERR_INVALID_DATE;

	exptime = 0.0;
	if (cmpack_ccd_get_exptime(ccdfile, &exptime)!=0)
		return CMPACK_ERR_READ_ERROR;
	if (exptime<=0.0) 
		return CMPACK_ERR_INVALID_EXPTIME;

	if (is_debug(lc->con)) {
		sprintf(msg, "Data format  : %s", pixformat(bitpix));
		printout(lc->con, 1, msg);
		sprintf(msg, "Min. value   : %.12g", minvalue); 
		printout(lc->con, 1, msg);
		sprintf(msg, "Max. value   : %.12g", maxvalue); 
		printout(lc->con, 1, msg);
  		sprintf(msg, "Image size   : %d x %d pixels", nx, ny); 
  		printout(lc->con, 1, msg);
		sprintf(msg, "Date & time  : %04d-%02d-%02d %02d:%02d:%02d.%03d UT", 
			dt.date.year, dt.date.month, dt.date.day, dt.time.hour, dt.time.minute, dt.time.second, dt.time.milisecond); 
  		printout(lc->con, 1, msg);
		sprintf(msg, "Exposure     : %.2f s", exptime); 
  		printout(lc->con, 1, msg);
  		sprintf(msg, "Image shift  : %d, %d pixels", offsetx, offsety); 
  		printout(lc->con, 1, msg);
	}

	/* Convert jd to center of exposure */
	jd += (0.5*exptime/86400.0);

    if (lc->in_width==0 && lc->in_height==0) {
	    /* first image, sets the variables */
	    lc->in_width = nx;
	    lc->in_height = ny;
		lc->sjd = jd;
	  	lc->nframe = 1;
		lc->exptime = exptime;
		lc->in_bitpix = bitpix;
		ccd_copy_header(ccdfile, lc->header);
	} else {
		/* Next image */
		if (nx!=lc->in_width || ny!=lc->in_height) 
			return CMPACK_ERR_DIFF_SIZE_SRC;
		if (bitpix!=lc->in_bitpix)
			return CMPACK_ERR_DIFF_BITPIX_SRC;
    	lc->sjd += jd;
    	lc->exptime += exptime;
  		lc->nframe++;
	}

    /* Read image data */
    ccd = cmpack_malloc(nx*ny*sizeof(double));
	res = cmpack_ccd_get_image(ccdfile, CMPACK_BITPIX_DOUBLE, 0, 0, ccd);
    if (res!=0) {
		cmpack_free(ccd);
	    return CMPACK_ERR_READ_ERROR;
    }

	if (lc->nframe==1) {
	    /* first image, init accumulation buffer */
		lc->data = cmpack_malloc(nx*ny*sizeof(double));
	 	lc->stat = cmpack_malloc(nx*ny*sizeof(char));
	  	for (j=0;j<ny;j++) {
	      	for (i=0;i<nx;i++) {
	          	ii = i + offsetx;
		        jj = j + offsety;
		        if (ii<lc->border.left || jj<lc->border.top || ii>=nx-lc->border.right || jj>=ny-lc->border.bottom) {
		      		S(i,j) = -2;
	      	    } else {
	               	x = I(ii,jj); 
	              	if (x>=maxvalue) {
	              		S(i,j) = 1;
	           		} else 
	           		if (x<=minvalue) {
		            	S(i,j) = -1;
	            	} else {
		            	S(i,j) = 0;
		            	D(i,j) = x;
	            	}
	  	      	}
	      	}
      	}
  	} else {
	 	/* Add to the accumulation buffer */
	  	for (j=0;j<ny;j++) {
	      	for (i=0;i<nx;i++) {
		        if (S(i,j)==0) {
	          		ii = i + offsetx;
		        	jj = j + offsety;
			        if (ii<lc->border.left || jj<lc->border.top || ii>=nx-lc->border.right || jj>=ny-lc->border.bottom) {
			      		S(i,j) = -2;
		      	    } else {
			      	    x = I(ii,jj);
		              	if (x>=maxvalue) {
			            	S(i,j) = 1;
		               	} else
		              	if (x<=minvalue) {
			            	S(i,j) = -1;
		            	} else {
			            	D(i,j) += x;
		            	}
	            	}
	  	      	}
	      	}
	  	}
  	}
  	cmpack_free(ccd);
    return 0;
}

/* Computes the output image and writes it to the file */
int cmpack_kombine_close(CmpackKombine *lc)
{
  	int res, i, j, nx, ny, bitpix;
  	int underflow = 0, overflow = 0, overexposed = 0, badpixels = 0;
  	char msg[MAXLINE];
  	double val, exptime, minvalue, maxvalue;
	CmpackDateTime dt;

	/* Check context */
	if (!lc->outfile) 
		return CMPACK_ERR_NO_OUTPUT_FILE;

  	/* Check data */
  	if (lc->nframe<=0 || !lc->data || !lc->stat) {
		cmpack_unref(lc->outfile);
		lc->outfile = NULL;
	  	return CMPACK_ERR_NO_INPUT_FILES;
	}

	/* Check image size */
	nx = lc->in_width;
	ny = lc->in_height;
	if (nx<=0 || ny<=0) {
		cmpack_unref(lc->outfile);
		lc->outfile = NULL;
		return CMPACK_ERR_INVALID_SIZE;
	}

	if (lc->bitpix==CMPACK_BITPIX_AUTO) {
		/* The output frame have the same format as the first source frame */
		bitpix = lc->in_bitpix;
	} else {
		/* Use required image data format */
		bitpix = lc->bitpix;
	}
	if (!pixrange(bitpix, &minvalue, &maxvalue)) {
		/* Invalid image data format */
		cmpack_unref(lc->outfile);
		lc->outfile = NULL;
		return CMPACK_ERR_INVALID_BITPIX;
	}

	/* Compute header information */
    exptime = lc->exptime;
	cmpack_decodejd(lc->sjd/lc->nframe - (0.5*exptime/86400.0), &dt);

	/* Write header information */
	ccd_prepare(lc->outfile, nx, ny, bitpix);
	ccd_copy_header(lc->header, lc->outfile);
	ccd_set_exptime(lc->outfile, lc->exptime);
	ccd_set_datetime(lc->outfile, &dt);
	ccd_set_frames(lc->outfile, lc->nframe);
	ccd_set_origin(lc->outfile);
	ccd_set_pcdate(lc->outfile);

	if (is_debug(lc->con)) {
		sprintf(msg, "Data format  : %d", bitpix); 
		printout(lc->con, 1, msg);
		sprintf(msg, "Min. value   : %.12g", minvalue); 
		printout(lc->con, 1, msg);
		sprintf(msg, "Max. value   : %.12g", maxvalue); 
		printout(lc->con, 1, msg);
  		sprintf(msg,"No. of frames: %d frames" , lc->nframe); 
  		printout(lc->con, 1, msg);
		sprintf(msg, "Date & time  : %04d-%02d-%02d %02d:%02d:%02d.%03d UT", 
			dt.date.year, dt.date.month, dt.date.day, dt.time.hour, dt.time.minute, dt.time.second, dt.time.milisecond); 
  		printout(lc->con, 1, msg);
  		sprintf(msg,"Exposure     : %.2f s", exptime); 
  		printout(lc->con, 1, msg);
	}

  	/* Computation output image data */
  	for (j=0;j<ny;j++) {
      	for (i=0;i<nx;i++) {
          	/* Compute the output value */
          	val = D(i,j)/lc->nframe;
          	if (S(i,j)==-2) {
	          	D(i,j) = 0.0;
          	} else
          	if (S(i,j)==-1) {
	          	D(i,j) = minvalue;
	          	badpixels++;
          	} else
          	if (S(i,j)==1) {
	          	D(i,j) = maxvalue;
      			overexposed++;
          	} else
          	if (val < minvalue) {
	          	D(i,j) = minvalue;
	          	underflow++;
          	} else
          	if (val > maxvalue) {
	          	D(i,j) = maxvalue;
	          	overflow++;
          	} else {
	          	D(i,j) = val;
          	}
      	}
  	}

	/* Write image data */
	res = ccd_write_image(lc->outfile, CMPACK_BITPIX_DOUBLE, lc->data);

	/* Range check */
	if (overflow>0) {
		sprintf(msg,"An overflow has been occurred on %d of %d pixels during computation (max.=%.12g).", 
			overflow, nx*ny, maxvalue);
		printout(lc->con, 0, msg);
	}
	if (underflow>0) {
		sprintf(msg,"An underflow has been occurred on %d of %d pixels during computation (min.=%.12g).", 
			underflow, nx*ny, minvalue);
		printout(lc->con, 0, msg);
	}

	/* Clear buffers */
	ccd_clear_header(lc->header);
	cmpack_free(lc->data);
	lc->data = NULL;
	cmpack_free(lc->stat);
	lc->stat = NULL;
	cmpack_unref(lc->outfile);
	lc->outfile = NULL;
	return res;
}
