/**************************************************************

konvoes.c (C-Munipack project)
Conversion functions for OES Astro files
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <time.h>

#include "config.h"
#include "oesfile.h"
#include "konvoes.h"
#include "comfun.h"
#include "console.h"

/* Returns nonzero if the file is in oes format */
int oes_test(const char *block, size_t bytes)
{
	unsigned long *buf = (unsigned long*)block;
	return (bytes>64*4) && ((buf[62]==H_MAG1_OWN && buf[63]==H_MAG2_OWN) || (buf[62]==H_MAG1_INV && buf[63]==H_MAG2_INV));
}

/* Read parameters from SBIG file */
int oes_open(tHandle *handle, const char *filename, CmpackOpenMode flags)
{
	return oesopen((struct oesfile**)handle, filename);
}

/* Close file and free allocated memory */
void oes_close(tHandle handle)
{
	oesclos(handle);
}

/* Reads string value from header */
int oes_gets(tHandle handle, const char *key, char **val)
{
	if (oesgkys((struct oesfile*)handle, atoi(key), val)==0)
		return 0;
	else
		return CMPACK_ERR_KEY_NOT_FOUND;
}

char *oes_getmagic(tHandle handle)
{
	return cmpack_strdup("OES Astro file");
}

/* Reads integer value from header */
int oes_geti(tHandle handle, const char *key, int *value)
{
	if (oesgkyi((struct oesfile*)handle, atoi(key), value)==0)
		return 0;
	else
		return CMPACK_ERR_KEY_NOT_FOUND;
}

int oes_getsize(tHandle handle, int *width, int *height)
{
	int nx, ny;

	if (oesghpr((struct oesfile*)handle, &nx, &ny)==0) {
		if (width) *width = nx;
		if (height) *height = ny;
		return 0;
	} else {
		if (width) *width = 0;
		if (height) *height = 0;
		return CMPACK_ERR_INVALID_SIZE;
	}
}

/* Reads floating point value from header */
int oes_getd(tHandle handle, const char *key, double *value)
{
	if (oesgkyd((struct oesfile*)handle, atoi(key), value)==0)
		return 0;
	else
		return CMPACK_ERR_KEY_NOT_FOUND;
}

/* Get exposure duration */
int oes_getexptime(tHandle handle, double *val)
{
	if (oesgexp((struct oesfile*)handle, val)==0)
		return 0;
	else
		return CMPACK_ERR_KEY_NOT_FOUND;
}

/* Get date and time of observation */
int oes_getdatetime(tHandle handle, CmpackDateTime *dt)
{
	memset(dt, 0, sizeof(CmpackDateTime));
	if (oesgdat((struct oesfile*)handle, &dt->date.year, &dt->date.month, &dt->date.day, 
		&dt->time.hour, &dt->time.minute, &dt->time.second)==0) 
		return 0;
	else
		return CMPACK_ERR_KEY_NOT_FOUND;
}

/* Get CCD temperature */
int oes_getccdtemp(tHandle handle, double *val)
{
	if (oesgtem((struct oesfile*)handle, val)==0)
		return 0;
	else
		return CMPACK_ERR_KEY_NOT_FOUND;
}

static int copyheader(struct oesfile *src, fitsfile *fits)
{
	int sstat, dstat = 0;
	char datestr[64], timestr[64];
	double exptime, ccdtemp;
	CmpackDateTime dt;

	/* Set date and time of observation */
	memset(&dt, 0, sizeof(CmpackDateTime));
	sstat = oesgdat(src, &dt.date.year, &dt.date.month, &dt.date.day, 
		&dt.time.hour, &dt.time.minute, &dt.time.second);
	if (sstat==0) {
		sprintf(datestr, "%04d-%02d-%02d", dt.date.year, dt.date.month, dt.date.day);
		ffukys(fits,"DATE-OBS", (char*) datestr, "UT DATE OF START", &dstat);
		sprintf(timestr, "%02d:%02d:%02d.%03d", dt.time.hour, dt.time.minute, dt.time.second, dt.time.milisecond);
		ffukys(fits,"TIME-OBS", (char*) timestr, "UT TIME OF START", &dstat);
	}

	/* Other parameters */
	sstat = oesgexp(src, &exptime);
	if (sstat==0) ffpkyg(fits,"EXPTIME",exptime,2,"EXPOSURE IN SECONDS",&dstat);
	sstat = oesgtem(src, &ccdtemp);
	if (sstat==0) ffpkyg(fits,"TEMPERAT",ccdtemp,1,"TEMPERATURE IN DEGREES C",&dstat);

	return dstat;
}

/* Add string to history */
int oes_copyheader(tHandle handle, fitsfile *fits)
{
	int dstat = copyheader((struct oesfile*)handle, fits);
	return (dstat!=0 ? CMPACK_ERR_WRITE_ERROR : 0);
}

/* Converts header from oes file to FITS file  */
int oes_copy(tHandle handle, fitsfile *fits, CmpackBitpix bitpix, int flipx, int flipy, 
			 const CmpackBorder *border, CmpackConsole *con)
{
	long naxes[2];
	char msg[MAXLINE], origin[128], actdate[128];
	time_t	tsec;
	struct tm *actim;
	int y, sstat, dstat=0, len, nx, ny;
	double *image;

	/* Get image dimensions */
	sstat = oesghpr((struct oesfile*)handle, &nx, &ny);
	if (sstat!=0) {
		if (con)
			printerr(con, sstat);
		return CMPACK_ERR_READ_ERROR;
	}

	/* Output data format */
	if (bitpix==CMPACK_BITPIX_AUTO)
		bitpix = CMPACK_BITPIX_FLOAT;
	if (bitpix > 0 && bitpix < CMPACK_BITPIX_SSHORT) {
		printout(con, 0, "Please, use higher image data precision.");
		return CMPACK_ERR_INVALID_BITPIX;
	}

	/* Set image parameters */
	naxes[0] = nx; 
	naxes[1] = ny;
	ffphps(fits,bitpix,2,naxes,&dstat);
	if (dstat!=0) {
		if (con) {
			ffgerr(dstat, msg);
			printout(con, 0, msg);
		}
		return CMPACK_ERR_WRITE_ERROR;
	}

	/* Copy file header */
	dstat = copyheader((struct oesfile*)handle, fits);
	if (dstat!=0) {
		if (con) {
			ffgerr(dstat, msg);
			printout(con, 0, msg);
		}
		return CMPACK_ERR_WRITE_ERROR;
	}

	/* Origin */
	sprintf(origin, "%s %s", PACKAGE_NAME, VERSION);
	ffukys(fits, "ORIGIN", origin, "CREATOR NAME AND VERSION", &dstat);
	tsec = time(&tsec);
	actim = localtime(&tsec);
	sprintf(actdate,"%04d-%02d-%02d",actim->tm_year+1900,actim->tm_mon+1,actim->tm_mday);
 	ffukys(fits, "DATE", actdate, "DATE OF CREATION", &dstat);

	/* Read image data */
	image = (double*)cmpack_calloc(nx*ny, sizeof(double));
	sstat = oesg2dd((struct oesfile*)handle,flipx,flipy,image);
	if (sstat!=0) { 
		if (con)
			printerr(con, sstat);
		cmpack_free(image); 
		return CMPACK_ERR_READ_ERROR;
	}

	/* Clear border */
	if (border->left>0) {
		len = (border->left < nx ? border->left : nx);
		for (y=0; y<ny; y++)
			memset(image+(y*nx), 0, len*sizeof(double));
	}
	if (border->right) {
		len = (border->right < nx ? border->right: nx);
		for (y=0; y<ny; y++)
			memset(image+((y+1)*nx-len), 0, len*sizeof(double));
	}
	if (border->top>0) {
		len = (border->top < ny ? border->top : ny);
		for (y=0; y<len; y++)
			memset(image+(y*nx), 0, nx*sizeof(double));
	}
	if (border->bottom>0) {
		len = (border->bottom < ny ? border->bottom : ny);
		for (y=0; y<len; y++)
			memset(image+((ny-y-1)*nx), 0, nx*sizeof(double));
	}

	/* Write image data */
	ffp2dd(fits,1,nx,nx,ny,image,&dstat);
	if (dstat!=0) { 
		if (con) {
			ffgerr(dstat, msg);
			printout(con, 0, msg);
		}
		cmpack_free(image); 
		return CMPACK_ERR_WRITE_ERROR; 
	}

	/* Normal return */
	cmpack_free(image);
	return 0;
}

/* Sequential read of parameters */
int oes_gkyn(tHandle handle, int nkey, char **key, char **val, char **com)
{
	return oesgkyn((struct oesfile*)handle, nkey, key, val, com);
}

/* Get image data */
int oes_getimage(tHandle handle, CmpackBitpix bitpix, int flipx, int flipy, void *buffer)
{
	switch (bitpix)
	{
	case CMPACK_BITPIX_SSHORT:
		return oesg2di((struct oesfile*)handle, 0, 0, (short*)buffer);
	case CMPACK_BITPIX_USHORT:
		return oesg2dui((struct oesfile*)handle, 0, 0, (unsigned short*)buffer);
	case CMPACK_BITPIX_SLONG:
		return oesg2dj((struct oesfile*)handle, 0, 0, (long*)buffer);
	case CMPACK_BITPIX_ULONG:
		return oesg2duj((struct oesfile*)handle, 0, 0, (unsigned long*)buffer);
	case CMPACK_BITPIX_FLOAT:
		return oesg2de((struct oesfile*)handle, 0, 0, (float*)buffer);
	case CMPACK_BITPIX_DOUBLE:
		return oesg2dd((struct oesfile*)handle, 0, 0, (double*)buffer);
	default:
		return CMPACK_ERR_INVALID_BITPIX;
	}
}

/* Get image data type */
CmpackBitpix oes_getbitpix(tHandle handle)
{
	return CMPACK_BITPIX_USHORT;
}

/* Get image data range */
int oes_getrange(tHandle handle, double *minvalue, double *maxvalue)
{
	if (minvalue)
		*minvalue = 0.0;
	if (maxvalue)
		*maxvalue = 0.0;
	return 0;
}
