/**************************************************************

list.c (C-Munipack project)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "comfun.h"
#include "console.h"
#include "hash.h"
#include "cmpack_common.h"
#include "cmpack_phtfile.h"
#include "cmpack_list.h"
#include "cmpack_table.h"
#include "diffmag.h"
#include "instmag.h"
#include "trklist.h"
#include "stddev.h"
#include "lister.h"

/*********************   LOCAL FUNCTIONS   ****************************/

/* Clear the listing context */
static void list_clean(void *ctx);

static CmpackLister *list_create(void)
{
	CmpackLister *pc = gc_create(sizeof(CmpackLister));
	pc->gc.clean = &list_clean;
	pc->aperture = 1;
	list_st_init(&pc->var);
	list_st_init(&pc->comp);
	list_st_init(&pc->chk);
	return pc;
}

static int list_init(CmpackLister *pc)
{
	switch (pc->format) 
	{
	case CMPACK_LIST_DIFFMAG:
		return list_diff_init(pc);
	case CMPACK_LIST_INSTMAG:
		return list_inst_init(pc);
	case CMPACK_LIST_TRACKLST:
		return list_trk_init(pc);
	case CMPACK_LIST_STDDEV:
		return list_dev_init(pc);
	default:
		printout(pc->con, 0, "Unknown output format specification");
		return CMPACK_ERR_INVALID_PAR;
	}
}

/* Clear the listing context */
static void list_clean(void *ctx)
{
	CmpackLister *lc = (CmpackLister*)ctx;

	list_st_clear(&lc->var);
	list_st_clear(&lc->comp);
	list_st_clear(&lc->chk);
	cmpack_free(lc->filter);
	cmpack_free(lc->filename);
	cmpack_unref(lc->tab);
	cmpack_unref(lc->con);
	gc_clean(ctx);
}

/********************   PUBLIC FUNCTIONS   **************************/

/* Make new listing context */
CmpackLister *cmpack_list_init(void)
{
	return list_create();
}

/* Attach console */
void cmpack_list_set_console(CmpackLister *ctx, CmpackConsole *con)
{
	if (con!=ctx->con) {
		cmpack_unref(ctx->con);
		ctx->con = cmpack_addref(con);
	}
}

/* Open output table, start reading sequence */
int cmpack_list_open(CmpackLister *pc, CmpackTable *tab, CmpackListFormat format, int flags)
{	
	int res;

	pc->format = format;
	pc->flags = flags;
	pc->tab = cmpack_addref(tab);
	cmpack_tab_clear(pc->tab);
	pc->filename = NULL;

	res = list_init(pc);
	if (res!=0) {
		cmpack_unref(pc->tab);
		pc->tab = NULL;
	}
	return res;
}

/* Set aperture index */
void cmpack_list_set_aperture(CmpackLister *proc, int index)
{
	proc->aperture = index;
}

/* Get aperture index */
int cmpack_list_get_aperture(CmpackLister *proc)
{
	return proc->aperture;
}

/* Set list of variable stars */
void cmpack_list_set_var(CmpackLister *lc, const int *items, int nitems)
{
	list_st_set(&lc->var, items, nitems);
}

/* Set list of comparison stars */
void cmpack_list_set_comp(CmpackLister *lc, const int *items, int nitems)
{
	list_st_set(&lc->comp, items, nitems);
}

/* Set list of check stars */
void cmpack_list_set_check(CmpackLister *lc, const int *items, int nitems)
{
	list_st_set(&lc->chk, items, nitems);
}

/* Reads magnitudes and error of the selected stars and writes them to output file */
int cmpack_list_read(CmpackLister *lc, CmpackPhtFile *pht, int frame)
{
	int res;

	if (!pht) {
		printout(lc->con, 0, "Invalid photometry file context.");
		return CMPACK_ERR_INVALID_PAR;
	}
	
	/* Read input file */
	switch (lc->format) 
	{
	case CMPACK_LIST_DIFFMAG:
		res=list_diff(lc, pht, frame);
		break;
	case CMPACK_LIST_INSTMAG:
		res=list_inst(lc, pht, frame);
		break;
    case CMPACK_LIST_TRACKLST:
		res=list_trk(lc, pht, frame);
      	break;
	case CMPACK_LIST_STDDEV:
		res = list_dev(lc, pht, frame);
		break;
	default:
		res = CMPACK_ERR_INVALID_PAR;
	}
	lc->nframes++;
	return res;
}

/* Finish reading */
int cmpack_list_close(CmpackLister *lc)
{
	int res;

	switch (lc->format) 
	{
	case CMPACK_LIST_DIFFMAG:
		res = list_diff_close(lc);
		break;
	case CMPACK_LIST_INSTMAG:
		res = list_inst_close(lc);
		break;
    case CMPACK_LIST_TRACKLST:
		res = list_trk_close(lc);
      	break;
	case CMPACK_LIST_STDDEV:
		res = list_dev_close(lc);
		break;
	default:
		res = CMPACK_ERR_INVALID_PAR;
	}

	if (res==0 && lc->tab && lc->filename) 
		cmpack_tab_save(lc->tab, lc->filename);

	cmpack_unref(lc->tab);
	lc->tab = NULL;
	cmpack_free(lc->filename);
	lc->filename = NULL;

	return res;
}

/* Get pointer to the associated table */
CmpackTable *cmpack_list_get_table(CmpackLister *proc)
{
	return proc->tab;
}
