/**************************************************************

aper.c (C-Munipack project)
Table of apertures (phootmetry file)
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "hash.h"
#include "comfun.h"
#include "console.h"
#include "cmpack_common.h"
#include "cmpack_phtfile.h"
#include "phtaper.h"

/*************************   Constants  ***********************/

#define ALLOC_BY 	64

/********************   Local functions    ***********************/

/* Init an aperture */
static CmpackPhtAperture *aperture_create(int aperid)
{
	CmpackPhtAperture *aper = gc_create(sizeof(CmpackPhtAperture));
	aper->id = aperid;
	return aper;
}

/* Copy section */
static void aperture_copy(CmpackPhtAperture *dst, const CmpackPhtAperture *src)
{
	dst->id = src->id;
	dst->mask = src->mask;
	dst->radius = src->radius;
}

/********************   Public functions    ***********************/

/* Init table of apertures */
void pht_aptab_init(CmpackAperTab *tab)
{
	memset(tab, 0, sizeof(CmpackAperTab));
}	

/* Free allocated memory */
void pht_aptab_clear(CmpackAperTab *tab)
{
	int j;
	
	for (j=0; j<tab->count; j++)
		cmpack_unref(tab->list[j]);
	cmpack_free(tab->list);
	tab->list = NULL;
	tab->count = tab->capacity = 0;
}	

/* Make copy of header */
void pht_aptab_copy(CmpackAperTab *dst, const CmpackAperTab *src)
{
	int i;
	
	pht_aptab_clear(dst);
	if (src->count>0) {
		dst->capacity = dst->count = src->count;
		dst->list = (CmpackPhtAperture**) cmpack_calloc(1, dst->capacity*sizeof(CmpackPhtAperture*));
		for (i=0; i<src->count; i++) {
			dst->list[i] = aperture_create(src->list[i]->id);
			aperture_copy(dst->list[i], src->list[i]);
		}
	}
}	

/* Dump table */
void pht_aptab_dump(CmpackConsole *con, CmpackAperTab *tab)
{
	int i;
	char aux[MAXLINE];
	
	sprintf(aux, "\tApertures: %d items", tab->count);
	printout(con, 1, aux);
	for (i=0; i<tab->count; i++) {
		CmpackPhtAperture *aper = tab->list[i];
		sprintf(aux, "\t\tAperture #%d", aper->id);
		printout(con, 1, aux);
		if (aper->mask & CMPACK_MASK_RADIUS) {
			sprintf(aux, "\t\t\tRadius: %.3f", aper->radius);
			printout(con, 1, aux);
		}
	}
}

/* Write table */
void pht_aptab_write(CmpackAperTab *tab, FILE *to)
{
	int i;
	
	for (i=0; i<tab->count; i++) {
		CmpackPhtAperture *aper = tab->list[i];
		fprintf(to, "\t<aper id=\"%d\"", aper->id);
		if (aper->mask & CMPACK_MASK_RADIUS) 
			fprintf(to, " radius=\"%.3f\"", aper->radius);
		fprintf(to, "/>\n");
	}
}

/* Add new section */
int pht_aptab_add(CmpackAperTab *tab, CmpackPhtAperture *aperture)
{
	int i = tab->count;
	if (tab->count>=tab->capacity) {
		tab->capacity += ALLOC_BY;
		tab->list = (CmpackPhtAperture**) cmpack_realloc(tab->list, tab->capacity*sizeof(CmpackPhtAperture*));
	}
	tab->list[i] = aperture;
	tab->count++;
	return i;
}

CmpackPhtAperture *pht_aper_new(int id)
{
	if (id>=0)
		return aperture_create(id);
	return NULL;
}

/* Delete a section */
void pht_aptab_delete(CmpackAperTab *tab, int index)
{
	CmpackPhtAperture *item = tab->list[index];
	/* Delete item */
	cmpack_unref(item);
	tab->count--;
	if (tab->count==0) {
		/* If there is no item, clear the list */
		cmpack_free(tab->list);
		tab->list = NULL;
		tab->count = tab->capacity = 0;
	} else {
		/* Pack the list (move all following items to the left) */
		size_t len = tab->count - index;
		if (len>0) {
			CmpackPhtAperture **aux = (CmpackPhtAperture**) cmpack_malloc(len*sizeof(CmpackPhtAperture*));
			memcpy(aux, item+1, len*sizeof(CmpackPhtAperture*));
			memcpy(item, aux, len*sizeof(CmpackPhtAperture*));
			cmpack_free(aux);
		}
	}
}

/**************************   PUBLIC INTERFACE   ******************************/

/* Create new aperture */
CmpackPhtAperture *cmpack_pht_aperture_new(int id)
{
	return pht_aper_new(id);
}

/* Get aperture identifier */
int cmpack_pht_aperture_get_id(CmpackPhtAperture *aper)
{
	return aper->id;
}

/* Get aperture radius */
double cmpack_pht_aperture_get_radius(CmpackPhtAperture *aper)
{
	if (aper->mask & CMPACK_MASK_RADIUS)
		return aper->radius;
	else
		return -1.0;
}

/* Set aperture radius */
void cmpack_pht_aperture_set_radius(CmpackPhtAperture *aper, double radius)
{
	if (aper && radius>=0) {
		aper->mask |= CMPACK_MASK_RADIUS;
		aper->radius = radius;
	}
}
