/**************************************************************

varfind.c (C-Munipack project)
Varfind data
Copyright (C) 2003 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <math.h>
#include <float.h>

#include "comfun.h"
#include "console.h"
#include "robmean.h"
#include "quicksort.h"
#include "cmpack_common.h"
#include "cmpack_phtfile.h"
#include "cmpack_table.h"
#include "stddev.h"
#include "lister.h"

int list_dev_init(CmpackLister *lc)
{
	int i, j;
   	char buf[MAXLINE];

	cmpack_tab_set_type(lc->tab, CMPACK_TABLE_APERTURES);

	/* Check parameters */
	if (lc->comp.count<1) {
		printout(lc->con, 0, "At least one comparison star must be selected.");
		return CMPACK_ERR_INVALID_PAR;
	}
	if (lc->chk.count<1) {
		printout(lc->con, 0, "At least one check star must be selected.");
		return CMPACK_ERR_INVALID_PAR;
	}

	/* Aperture */
	cmpack_tab_add_column(lc->tab, "APERTURE", CMPACK_COLUMN_INTEGER, 0);

	/* C-K1, C-K2, ... */
	for (i=0; i<lc->chk.count; i++) {
		sprintf(buf, "C-K%d", i+1);
		cmpack_tab_add_column(lc->tab, buf, CMPACK_COLUMN_DOUBLE, MAG_PRECISION);
	}

	/* K1-K2, ... */
	for (i=0; i<lc->chk.count; i++) {
		for (j=i+1; j<lc->chk.count; j++) {
			sprintf(buf, "K%d-K%d", i+1, j+1);
			cmpack_tab_add_column(lc->tab, buf, CMPACK_COLUMN_DOUBLE, MAG_PRECISION);
		}
	}

	return 0;
}

int list_dev(CmpackLister *lc, CmpackPhtFile *pht, int frame)
{
	int		ok, i, j, k, acount, scount, xref, ap_id;
	char	msg[MAXLINE];
	const char *filter;
	double  radius;
	CmpackPhtAperture *aper;
	CmpackPhtStar *st;
	CmpackPhtMeasurement data;
	CmpackStddevMag *dta;
	CmpackStddevStar *star;
	CmpackStar *comp, *chk, *in_st;
	CmpackStddevFrameList *flist = &lc->dflist;
	CmpackStddevFrame *ptr;
	CmpackStddevAperList *aplist = &lc->dalist;

	/* Get optical filter name */
	filter = cmpack_pht_get_filter(pht);
	/* Check parameters */
	if (lc->nframes==0) {
		if (filter) {
			lc->filter = cmpack_strdup(filter);
			if (strlen(lc->filter)>0)
				cmpack_tab_pkys(lc->tab, "Filter", lc->filter);
		}
	} else {
		if (filter) {
			if (strcmp(lc->filter, filter)!=0)
				printout(lc->con, 0, "Different optical filter!");
		} else {
			if (lc->filter)
				printout(lc->con, 0, "Different optical filter!");
		}
	}

	/* Create a new frame */
	if (flist->count>=flist->capacity) {
		flist->capacity += 64;
		flist->list = (CmpackStddevFrame*) cmpack_realloc(flist->list, flist->capacity*sizeof(CmpackStddevFrame));
	}
	ptr = &flist->list[flist->count++];
	memset(ptr, 0, sizeof(CmpackStddevFrame));

	/* Check aperture */
	acount = cmpack_pht_naper(pht);
	for (k=0; k<acount; k++) {
		aper = cmpack_pht_get_aperture(pht, k);
		ap_id = cmpack_pht_aperture_get_id(aper);
		if (ap_id>0) {
			radius = cmpack_pht_aperture_get_radius(aper);
			ok = 0;
			for (i=0; i<aplist->count; i++) {
				if (aplist->list[i].id==ap_id) {
					if (radius!=aplist->list[i].radius) {
						sprintf(msg, "Aperture #%d: Different aperture radius!", ap_id);
						printout(lc->con, 0, msg);
					}
					ok = 1;
					break;
				}
			}
			if (!ok) {
				/* Add new aperture */
				if (aplist->count>=aplist->capacity) {
					aplist->capacity += 16;
					aplist->list = cmpack_realloc(aplist->list, aplist->capacity*sizeof(CmpackStddevAperture));
				}
				aplist->list[aplist->count].id = ap_id;
				aplist->list[aplist->count].radius = radius;
				aplist->count++;
			}
		}
		cmpack_unref(aper);
	}

	/* Read magnitudes */
	scount = cmpack_pht_nstar(pht);
	acount = cmpack_pht_naper(pht);
	comp = lc->comp.list;
	chk = lc->chk.list;
	for (i=0; i<scount; i++) {
		st = cmpack_pht_get_star(pht, i);
		xref = cmpack_pht_star_get_refid(st);
		if (xref>=0) {
			in_st = NULL; 
			for (j=0; j<lc->comp.count && !in_st; j++)
				if (lc->comp.list[j].id==xref) in_st=&comp[j];
			for (j=0; j<lc->chk.count && !in_st; j++)
				if (lc->chk.list[j].id==xref) in_st=&chk[j];
			if (in_st) {
				star = (CmpackStddevStar*) cmpack_calloc(1, sizeof(CmpackStddevStar));
				if (star) {
					star->id = xref;
					for (k=0; k<acount; k++) {
						aper = cmpack_pht_get_aperture(pht, k);
						ap_id = cmpack_pht_aperture_get_id(aper);
						if (cmpack_pht_star_get_measurement(st, ap_id, &data)==0) {
							if (data.mask & CMPACK_MASK_MAG) {
								/* Add measurement to image data */
								dta = (CmpackStddevMag*) cmpack_calloc(1, sizeof(CmpackStddevMag));
								if (dta) {
									dta->id = ap_id;
									dta->mag = data.magnitude;
									if (!star->last) {
										star->first = dta;
										star->last = dta; 
									} else { 
										star->last->next = dta; 
										star->last = dta; 
									}
								}
							}
						}
						cmpack_unref(aper);
					}
				}
				if (!ptr->last) {
					ptr->first = star;
					ptr->last = star;
				} else {
					ptr->last->next = star;
					ptr->last = star;
				}
			}
		}
		cmpack_unref(st);
	}
	/* Normal return */
	return 0; 
}

int list_dev_close(CmpackLister *lc)
{
	int	a, i, j, k, col, row, count, n, ls_ncomp, ls_nchk, ls_nimage;
	char *ls_cok;
	double cval, *ls_dmag, *ls_cmag, mean, stddev;
	CmpackStddevAperture *aper;
	CmpackStddevMag **ls_comp, **ls_chk;
	CmpackStddevMag *mag, *comp, *chk, *chk1, *chk2;
	CmpackStddevFrame *image;
	CmpackStddevStar *star;

	ls_nimage   = lc->dflist.count;
	ls_ncomp	= lc->comp.count;
	ls_nchk		= lc->chk.count;
	if (ls_nimage<=0)
		return CMPACK_ERR_NO_INPUT_FILES;

    ls_comp		= (CmpackStddevMag**) cmpack_malloc(ls_ncomp*ls_nimage*sizeof(CmpackStddevMag*));
	ls_chk		= (CmpackStddevMag**) cmpack_malloc(ls_nchk*ls_nimage*sizeof(CmpackStddevMag*));
	ls_dmag     = (double*) cmpack_malloc(ls_nimage*sizeof(double));
	ls_cmag     = (double*) cmpack_malloc(ls_nimage*sizeof(double));
	ls_cok      = (char*) cmpack_malloc(ls_nimage*sizeof(char));

	/* For all apertures */
	for (a=0; a<lc->dalist.count; a++) {
		aper = &lc->dalist.list[a];
		row = cmpack_tab_add_row(lc->tab);
		cmpack_tab_ptdi(lc->tab, row, 0, aper->id);
		memset(ls_comp, 0, ls_ncomp*ls_nimage*sizeof(CmpackStddevMag*));
		memset(ls_chk, 0, ls_nchk*ls_nimage*sizeof(CmpackStddevMag*));
		/* For all images */
		for (i=0; i<lc->dflist.count; i++) {
			image = &lc->dflist.list[i]; 
			/* For all comparison stars */
			for (j=0; j<ls_ncomp; j++) {
				star = image->first;
				while (star) {
					if (star->id==lc->comp.list[j].id) {
						mag = star->first;
						while (mag) {
							if (mag->id==aper->id) {
								ls_comp[i+j*ls_nimage] = mag;
								break;
							}
							mag = mag->next;
						}
						break;
					}
					star = star->next;
				}
			}
			/* For all check stars */
			for (j=0; j<ls_nchk; j++) {
				star = image->first;
				while (star) {
					if (star->id==lc->chk.list[j].id) {
						mag = star->first;
						while (mag) {
							if (mag->id==aper->id) {
								ls_chk[i+j*ls_nimage] = mag;
								break;
							}
							mag = mag->next;
						}
						break;
					}
					star = star->next;
				}
			}
		}
		/* Compute artifitial comparison star magnitude */
		memset(ls_cok, 0, ls_nimage*sizeof(char));
		for (i=0; i<ls_nimage; i++) {
			cval = 0.0; n = 0;
			for (j=0; j<ls_ncomp; j++) {
				comp = ls_comp[i+j*ls_nimage];
				if (comp) {
					cval += pow(10.0, -0.4*comp->mag);
					n++;
				}
			}
			if (n!=ls_ncomp) {
				printout(lc->con, 1, "The comparison star(s) have incorrect value of magnitude");
			} else {
				ls_cmag[i] = -2.5*log10(cval/n);
				ls_cok[i] = 1;
			}
		}
		/* Compute standard errors */
		col = 1;
		/* C-C1, C-C2, ... */
		for (j=0; j<ls_nchk; j++) {
			count = 0;
			for (i=0; i<ls_nimage; i++) {
				chk = ls_chk[i+j*ls_nimage];
				if (chk && ls_cok[i])
					ls_dmag[count++] = ls_cmag[i] - chk->mag;
			}
			if (count>1) {
				cmpack_robustmean(count, ls_dmag, &mean, &stddev);
				cmpack_tab_ptdd(lc->tab, row, col, stddev);
			} else {
				cmpack_tab_ptdd(lc->tab, row, col, 9.9999);
			}
			col++;
		}
		/* C1-C2, ... */
		for (j=0; j<ls_nchk; j++) {
			for (k=j+1; k<ls_nchk; k++) {
				count = 0;
				for (i=0; i<ls_nimage; i++) {
					chk1 = ls_chk[i+j*ls_nimage];
					chk2 = ls_chk[i+k*ls_nimage];
					if (chk1 && chk2)
						ls_dmag[count++] = chk1->mag - chk2->mag;
				}
				if (count>1) {
					cmpack_robustmean(count, ls_dmag, &mean, &stddev);
					cmpack_tab_ptdd(lc->tab, row, col, stddev);
				} else {
					cmpack_tab_ptdd(lc->tab, row, col, 9.9999);
				}
				col++;
			}
		}
	}

	for (i=0; i<lc->dflist.count; i++) {
		star = lc->dflist.list[i].first;
		while (star) {
			CmpackStddevStar *next_star = star->next;
			mag = star->first;
			while (mag) {
				CmpackStddevMag *next_mag = mag->next;
				cmpack_free(mag);
				mag = next_mag;
			}
			cmpack_free(star);
			star = next_star;
		}
	}

	if (lc->comp.list) {
		cmpack_free(lc->comp.list);
		lc->comp.list = NULL;
		lc->comp.count = 0;
	}
	if (lc->chk.list) {
		cmpack_free(lc->chk.list);
		lc->chk.list = NULL;
		lc->chk.count = 0;
	}
	if (lc->dflist.list) {
		cmpack_free(lc->dflist.list);
		lc->dflist.list = NULL;
		lc->dflist.count = 0;
	}
	if (lc->dalist.list) {
		cmpack_free(lc->dalist.list);
		lc->dalist.list = NULL;
		lc->dalist.count = 0;
	}
	cmpack_free(ls_comp);
	cmpack_free(ls_chk);
	cmpack_free(ls_dmag);
	cmpack_free(ls_cmag);
	cmpack_free(ls_cok);
	return 0;
}
