/**************************************************************

chart.h (C-Munipack project)
Widget which can draw a chart
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#ifndef MUNIWIN_CHART_H
#define MUNIWIN_CHART_H

#include <gtk/gtk.h>
#include <cmunipack.h>

#include "callback.h"
#include "helper_classes.h"

class CChartBase:public CCBObject
{
public:
	// Star parameters
	struct tStarInfo
	{
		int			id;			// Star identifier
		double		x, y;		// Position
		bool		matched;	// Star has been matched
		int			xref;		// Reference id
		tStarType	sel_type;	// Selection type
		int			sel_index;	// Selection index
		bool		mag_valid;	// Magnitude is valid
		double		mag;		// Brightness in current aperture
		bool		err_valid;	// Error estimation is valid
		double		err;		// Error estimation
	};

	// Aperture parameters
	struct tAperInfo
	{
		int			id;			// Aperture identifier
		double		radius;		// Radius in pixels
	};

	enum tMessageId
	{
		CB_LEFT_CLICK,			// Left button click
		CB_LEFT_DBLCLICK,		// Left button double click
		CB_RIGHT_CLICK,			// Right button click
		CB_MOUSE_MOVE,			// Mouse moved
		CB_ZOOM_CHANGE			// Zoom factor changed
	};
	
public:
	// Constructor
	CChartBase();

	// Destructor
	virtual ~CChartBase();

	// Clear all data
	void Clear(void);

	// Load image 
	void LoadImage(const char *filename);

	// Load chart
	void LoadChart(const char *filename, bool ignore_selection);

	// Set selection
	void SetSelection(const CSelection *sel);

	// Clear selection of stars
	CSelection *GetSelection(void)
	{ return &m_Selection; }

	// Enable/disable mouse tracking
	void EnableTracking(bool enable);

	// Set pointer to the grayscale
	void SetScale(class CScale *pScale);

	// Is image inverted?
	bool ImageInvert(void) const
	{ return m_Invert; }

	// Turn pseudocolors on/off
	void SetPseudocolors(bool enable);

	// Are pseudocolors enabled?
	bool Pseudocolors(void) const
	{ return m_Pseudocolors; }

	// Show bad pixels
	void SetBadPixels(bool show);

	// Bad pixels?
	bool BadPixels(void) const
	{ return m_BadPixels; }

	// Show/hide image
	void ShowImage(bool show);

	// Show/hide chart
	void ShowChart(bool show);

	// Show/hide bad stars (invalid magnitude)
	void ShowBadStars(bool show);

	// Show bad stars?
	bool BadStars(void) const
	{ return m_BadStars; }

	// Show/hide not matched stars
	void ShowNonMatchedStars(bool show);

	// Show not matched stars?
	bool NonMatchedStarsVisible(void) const
	{ return m_ShowNonMatchedStars; }

	// Show/hide not matched stars
	void HighlightNonMatchedStars(bool show);

	// Show not matched stars?
	bool NonMatchedStarsHighlighted(void) const
	{ return m_HighlightNonMatchedStars; }

	// Show/hide not matched stars
	void HighlightSelectedStars(bool show);

	// Show not matched stars?
	bool SelectedStarsHighlighted(void) const
	{ return m_HighlightSelectedStars; }

	// Set active aperture
	void SetAperture(int aperture);

	// Paint chart to a drawing context
	void Export(cairo_surface_t *cr, int width, int height) const;

	// Get scale min and max 
	void GetScaleMinMax(double *min, double *max);

	// Get cursor position (in chart coordinates)
	bool GetCursorPos(int *x, int *y) const;

	// Get pixel value
	double GetValue(int x, int y) const;

	// Find star by identifier
	int FindStarById(int id);

	// Find star by identifier
	int FindStarByRefId(int xref);

	// Find star by position
	int FindStarByPos(int x, int y, bool visible_only);

	// Set focused star
	void SetFocusedStar(int index);

	// Get focused star
	int FocusedStar() const
	{ return m_StFocused; }

	// Select star
	void SelectStar(int index);

	// Get selected star
	int SelectedStar() const
	{ return m_StSelected; }

	// Get star data
	bool GetStar(int index, tStarInfo *info) const;

	// Quick photometry
	bool Photometry(int x, int y, double aperture, struct tPhotResult *result);

	// Get image information
	const CImageInfo *ImageInfo(void) const
	{ return &m_CCDInfo; }

	// Get chart information
	const CChartInfo *ChartInfo(void) const
	{ return &m_PhtInfo; }

protected:
	// Measurement
	struct tMagRec
	{
		int mask;
		double mag, err;
	};

	// Star parameters
	struct tStarRec
	{
		int			id;			// Star identifier
		double		x, y;		// Position
		bool		matched;	// Star was matched
		int			xref;		// Reference id
		tMagRec		*data;		// Measurements
		tStarType	stype;		// Star type
	};

	// Aperture parameters
	struct tAperRec
	{
		int Id;					// Aperture identifier
		double Radius;			// Radius in pixels
		double RefMag;			// Reference magnitude (for drawing stars)
	};

	// Image data
	struct tImageData
	{
		int		width, height;				// Image size
		double	scale, offset;				// Brightness scale and offset
		double *ccd;						// CCD data
		cairo_surface_t *image;				// Background image
		double	datamin, datamax;			// Allowed data range
	};

	// Photometry data
	struct tChartData
	{
		int			width, height;			// Chart size
		tAperRec	*aper;					// List of apertures
		int			ap_count;				// Number of apertures, allocated space
		tStarRec	*stars;					// List of stars
		int			st_count;				// Number of stars, allocated space
	};

	GtkWidget			*m_DrawArea;				// Drawing area
	GdkPixmap			*m_Pixmap;					// Backscreen buffer
	bool				m_NeedUpdate;				// Redraw backscreen buffer at next paint
	int					m_PixWidth, m_PixHeight;	// Backscreen buffer size
	int					m_ApertureID;				// Id of active aperture 
	int					m_ApertureIndex;			// Index of active aperture
	int					m_StFocused;				// Index of focused star
	int					m_StSelected;				// Selected star
	bool				m_EnableTrack;				// Mouse tracking enabled?
	bool				m_ShowImage;				// Show CCD frame?
	bool				m_ShowChart;				// Show stars?
	bool				m_BadStars;					// Show bad stars
	bool				m_Pseudocolors;				// Show pseudocolors
	bool				m_BadPixels;				// Show bad pixels
	bool				m_ShowNonMatchedStars;		// Show non-matched stars
	bool				m_HighlightNonMatchedStars;	// Highlight non-matched stars
	bool				m_HighlightSelectedStars;	// Highlight selected stars
	bool				m_Invert;					// Positive/negative image
	int					m_CursorX, m_CursorY;		// Last cursor position
	tImageData			m_CCDData;					// CCD image data
	CImageInfo			m_CCDInfo;					// CCD image info
	tChartData			m_PhtData;					// Photometry file data
	CChartInfo			m_PhtInfo;					// Photometry file info
	CScale				*m_pScale;					// Pointer to the scale box
	CSelection			m_Selection;				// Selected stars
	double				m_Zoom;						// Zoom between view and image
	double				m_OffsetX, m_OffsetY;		// Offset between view and image
	int					m_MapWidth, m_MapHeight;	// Reference size (image size or chart size)
	cairo_surface_t		*m_Image;					// Background image
	int					m_ImgWidth, m_ImgHeight;	// Background image size
	bool				m_IgnoreApertures;			// Ignore apertures

	// Mapping functions
	double XToView(double x) const;
	double YToView(double y) const;
	double ViewToX(double u) const;
	double ViewToY(double v) const;

	// Load photometry file
	void LoadPhotometryFile(CmpackPhtFile *file);

	// Load catalogue file
	void LoadCatalogueFile(CmpackCatFile *file);

	// After this call, the backscreen buffer will be updated before next paint.
	void Invalidate(void);

	// Update back-screen buffer
	virtual void Redraw(void);

	// Draw to buffer
	void Draw(cairo_t *cr, double zoom, double offsetx, double offsety) const;

	// Draw star
	void DrawStar(cairo_t *cr, double zoom, double offsetx, double offsety, 
		const tStarRec *star, bool selected, bool invert) const;

	// Draw label
	void DrawLabel(cairo_t *cr, double zoom, double offsetx, double offsety, 
		const tStarRec *star, bool selected, bool invert) const;

	// Mouse move tracking
	void MouseMove(GdkModifierType state, int x, int y);

	// Clear photometry data
	void ClearPhotometry(void);

	// Clear image data
	void ClearImage(void);

	// Clear selection
	void ClearSelection(void);

	// Compute circle radius for given star
	double ComputeRadius(const tStarRec *star, double zoom) const;

	// Resize drawing area
	virtual void Resize(void);

	// Rebuild image
	void RebuildImage(void);

	// Update star types
	void UpdateTypes(void);

	// Find aperture index
	int FindApIndex(int aperture) const;

	virtual void OnImageSizeChanged(void) {};

	// Create a new backing pixmap of the appropriate size 
	static gint configure_event(GtkWidget *widget, GdkEventConfigure *event, CChartBase *pMe);
	void OnConfigure(GtkWidget *widget, GdkEventConfigure *event);

	// Redraw the screen from the backing pixmap 
	static gint expose_event(GtkWidget *widget, GdkEventExpose *event, CChartBase *pMe);
	void OnExposeEvent(GtkWidget *widget, GdkEventExpose *event);

	// Click handling
	static gint button_press_event(GtkWidget *widget, GdkEventButton *event, CChartBase *pMe);
	void OnButtonPress(GtkWidget *widget, GdkEventButton *event);

	// Mouse cursor tracking
	static gint motion_notify_event(GtkWidget *widget, GdkEventMotion *event, CChartBase *pMe);
	void OnMotionNotify(GtkWidget *widget, GdkEventMotion *event);

	// Mouse leave tracking
	static gboolean leave_notify_event(GtkWidget *widget, GdkEventCrossing *event, CChartBase *pMe);
	void OnLeaveNotify(GtkWidget *widget, GdkEventCrossing *event);

	// Compare apertures by id
	static gint compare_apers(const CChartBase::tAperRec *a, const CChartBase::tAperRec *b);

	// Compare stars by magnitude
	static gint compare_stars(const tStarRec *a, const tStarRec *b);

	// Selection callback
	static void SelectionCallback(int message, int wparam, void* lparam, void* cb_data);
};

class CChart:public CChartBase
{
public:
	// Constructor
	CChart();

	// Get handle to the drawing area
	GtkWidget *Handle(void)
	{ return m_DrawArea; }

private:
	// Update back-screen buffer
	virtual void Redraw(void);
};

#endif
