/**************************************************************

deviations_dlg.cpp (C-Munipack project)
The 'Plot deviations' dialog
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <cmunipack.h>

#include "project.h"
#include "configuration.h"
#include "graph.h"
#include "progress_dlg.h"
#include "deviations_dlg.h"
#include "main.h"
#include "utils.h"
#include "proc_classes.h"

//-------------------------   LIST COLUMNS   --------------------------------

struct tTreeViewColumn {
	const char *caption;		// Column name
	int column;					// Model column index
	GtkTreeCellDataFunc datafn;	// Data function
	gfloat align;				// Text alignment
};

enum tDataColumnId {
	DATA_COLUMN,				// Column ordinal number
	DATA_NAME,					// Column name
	DATA_NCOLS					// Number of columns
};

const static tTreeViewColumn DataListColumns[] = {
	{ "Data",			DATA_NAME,	NULL,	0 },
	{ NULL }
};

//-------------------------   CHOOSE STARS DIALOG   --------------------------------

CDeviationsDlg::CDeviationsDlg(GtkWindow *pParent):m_Column(0)
{
	int i;
	GtkWidget *frame, *hbox, *vbox, *scrwnd;

	// Create table context
	m_Table = cmpack_tab_init(CMPACK_TABLE_APERTURES);

	// Dialog with buttons
	m_pDlg = gtk_dialog_new_with_buttons("Plot deviations", pParent, 
		(GtkDialogFlags)(GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR),
		GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE, GTK_STOCK_SAVE, GTK_RESPONSE_NONE, NULL);

	// Dialog icon
	gchar *icon = get_icon_file("deviations");
	gtk_window_set_icon(GTK_WINDOW(m_pDlg), gdk_pixbuf_new_from_file(icon, NULL));
	g_free(icon);

	// Dialog layout
	hbox = gtk_hbox_new(FALSE, 8);
	gtk_box_pack_start(GTK_BOX(GTK_DIALOG(m_pDlg)->vbox), hbox, TRUE, TRUE, 0);
	vbox = gtk_vbox_new(FALSE, 8);
	gtk_box_pack_start(GTK_BOX(hbox), vbox, FALSE, TRUE, 0);

	// List of data columns
	m_DataList = gtk_list_store_new(DATA_NCOLS, G_TYPE_INT, G_TYPE_STRING);
	m_DataView = gtk_tree_view_new_with_model(GTK_TREE_MODEL(m_DataList));
	for (i=0; DataListColumns[i].caption!=NULL; i++) {
		GtkTreeViewColumn *col = gtk_tree_view_column_new();
		// Set column name and alignment
		gtk_tree_view_column_set_title(col, DataListColumns[i].caption);
		gtk_tree_view_append_column(GTK_TREE_VIEW(m_DataView), col);
		// Add text renderer
		GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
		gtk_tree_view_column_pack_start(col, renderer, TRUE);
		g_object_set(renderer, "xalign", DataListColumns[i].align, NULL);
		if (DataListColumns[i].datafn)
			gtk_tree_view_column_set_cell_data_func(col, renderer, DataListColumns[i].datafn, NULL, NULL);
		else
			gtk_tree_view_column_add_attribute(col, renderer, "text", DataListColumns[i].column);
	}
	gtk_tree_selection_set_mode(gtk_tree_view_get_selection(GTK_TREE_VIEW(m_DataView)), GTK_SELECTION_SINGLE);
	gtk_widget_set_size_request(m_DataView, 140, 0);

	// Register callback for selection change
	GtkTreeSelection *selection = gtk_tree_view_get_selection(GTK_TREE_VIEW(m_DataView));
	gtk_tree_selection_set_mode(selection, GTK_SELECTION_BROWSE);
	g_signal_connect(G_OBJECT(selection), "changed", G_CALLBACK(selection_changed), this);

	// Make scrolled window
	scrwnd = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrwnd),
		GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(scrwnd), 
		GTK_SHADOW_ETCHED_IN);
	gtk_container_add(GTK_CONTAINER(scrwnd), m_DataView);
	gtk_box_pack_start(GTK_BOX(vbox), scrwnd, TRUE, TRUE, 0);

	// Graph
	frame = gtk_frame_new(NULL);
	gtk_box_pack_start(GTK_BOX(hbox), frame, true, true, 0);
	gtk_container_add(GTK_CONTAINER(frame), GTK_WIDGET(g_object_ref(m_Graph.Handle())));
	gtk_widget_set_size_request(m_Graph.Handle(), 640, 480);
	gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_IN);
	m_Graph.SetData(&m_Data);
	m_Graph.SetGraphPadding(2, 2, 48, 2);
	m_Graph.EnableTracking(false);

	gtk_widget_show_all(hbox);

	// Signals:
	g_signal_connect(G_OBJECT(m_pDlg), "response", G_CALLBACK(response_dialog), this);
}

CDeviationsDlg::~CDeviationsDlg()
{
	gtk_widget_destroy(m_pDlg);
	g_object_unref(m_DataList);
	cmpack_unref(m_Table);
}

void CDeviationsDlg::Execute()
{
	int i, index, j, res, col_count, row_count;
	double aper, *yvals;
	GtkWindow *pMainWnd = Global.pMainWnd->GetHandle();
	const CSelection *sel;
	GtkTreeIter iter;

	CProgressDlg::InitBuffer();
	cmpack_tab_clear(m_Table);
	m_Data.Clear();

	// Load reference file
	if (!Global.pProject->GetReference(NULL, NULL)) {
		ShowError(pMainWnd, "Please, execute the matching procedure first.");
		return;
	}

	// Check selection
	sel = Global.pProject->Selection();
	if (sel->CountStars(ST_COMPARISON)<=0 || sel->CountStars(ST_CHECK)<=0) {
		ShowError(pMainWnd, "Please, select at least one comparison star and at least one check star.");
		return;
	}

	// Make list of input files
	m_FileList = NULL;
	gtk_tree_model_foreach(Global.pProject->FileList(), (GtkTreeModelForeachFunc)foreach_all_files, &m_FileList);
	if (!m_FileList) {
		ShowError(pMainWnd, "There are no valid frames in the list.");
		return;
	}

	// Process files
	CProgressDlg pDlg(pMainWnd, "Processing files...");
	res = pDlg.Execute(ExecuteProc, this);

	g_slist_foreach(m_FileList, (GFunc)gtk_tree_row_reference_free, NULL);
	g_slist_free(m_FileList);
	m_FileList = NULL;

	if (pDlg.Cancelled())
		return;

	if (res!=0) {
		char *msg = cmpack_formaterror(res);
		ShowError(pMainWnd, msg);
		cmpack_free(msg);
		return;
	}

	// Initialize graph data
	m_Data.LockRedraw();
	m_Data.Clear();
	col_count = cmpack_tab_ncolumns(m_Table)-1;
	row_count = cmpack_tab_nrows(m_Table);
	if (col_count>0 && row_count>0) {
		m_Data.Init(col_count, row_count);
		yvals = (double*)g_malloc(col_count*sizeof(double));
		for (i=0; i<row_count; i++) {
			cmpack_tab_gtdd(m_Table, i, 0, &aper);
			for (j=0; j<col_count; j++)
				cmpack_tab_gtdd(m_Table, i, j+1, yvals+j);
			m_Data.Append(aper, yvals);
		}
	}
	m_Data.UnlockRedraw();

	// Set graph parameters
	m_Graph.SetXParams("Aperture", 1, MAX_APERTURES, MAX_APERTURES, 0, 0);
		
	// Initialize list of columns
	gtk_tree_view_set_model(GTK_TREE_VIEW(m_DataView), NULL);
	gtk_list_store_clear(m_DataList);
	m_Graph.ClearColumns();
	for (i=0; i<col_count; i++) {
		CmpackTabColumn *col = cmpack_tab_get_column(m_Table, i+1);
		if (col) {
			index = m_Graph.AddColumn(i, -1, 0);
			gtk_list_store_append(m_DataList, &iter);
			gtk_list_store_set(m_DataList, &iter, DATA_COLUMN, index, 
				DATA_NAME, cmpack_tab_column_get_name(col), -1);
			cmpack_unref(col);
		}
	}
	gtk_tree_view_set_model(GTK_TREE_VIEW(m_DataView), GTK_TREE_MODEL(m_DataList));

	SetData(0);
	m_Graph.AutoZoom(true, true);

	gtk_dialog_run(GTK_DIALOG(m_pDlg));
}

int CDeviationsDlg::ExecuteProc(CProgressDlg *sender, void *userdata)
{
	return ((CDeviationsDlg*)userdata)->ProcessFiles(sender);
}

int CDeviationsDlg::ProcessFiles(CProgressDlg *sender)
{
	int res, frameid, infiles;
	char *temp, msg[128];
	CListProc list;

	infiles = 0;
	sender->SetMinMax(0, g_slist_length(m_FileList));
	sender->Print("------ Table of apertures ------");

	res = list.Open(sender, m_Table, CMPACK_LIST_STDDEV);
	if (res==0) {
		for (GSList *node = m_FileList; node != NULL; node = node->next) {
			GtkTreePath *path = gtk_tree_row_reference_get_path((GtkTreeRowReference*)node->data);
			frameid = Global.pProject->GetFrameID(path);
			temp = Global.pProject->GetPhotFileName(path);
			sender->SetFileName(temp);
			g_free(temp);
			sprintf(msg, "Frame #%d:", frameid);
			sender->Print(msg);
			sender->SetProgress(infiles++);
			res = list.Add(path);
			if (res!=0 || sender->Cancelled()) 
				break;
		}
		res = list.Close();
	}
	if (res==0) {
		if (sender->Cancelled())
			sprintf(msg, "Cancelled at the user's request");
		else
			sprintf(msg, "====== %d succeeded ======", infiles);
		sender->Print(msg);
	} else {
		char *aux = cmpack_formaterror(res);
		sender->Print(msg);
		cmpack_free(aux);
	}
	return res;
}

gboolean CDeviationsDlg::foreach_all_files(GtkTreeModel *model, GtkTreePath *path, 
										GtkTreeIter *iter, GSList **list)
{
	if (Global.pProject->GetState(iter) & CFILE_MATCHING) 
		*list = g_slist_append(*list, gtk_tree_row_reference_new(model, path));
	return FALSE;
}

void CDeviationsDlg::SetData(int column)
{
	gchar *name;
	GtkTreeIter iter;

	m_Column = column;
	m_Graph.ShowColumn(-1, false);
	m_Graph.ShowColumn(m_Column, true); 
	if (gtk_tree_model_iter_nth_child(GTK_TREE_MODEL(m_DataList), &iter, NULL, m_Column)) {
		gtk_tree_model_get(GTK_TREE_MODEL(m_DataList), &iter, DATA_NAME, &name, -1);
		m_Graph.SetYParams(name, 0.0, 99.0, 0.1, 0, 1);
		g_free(name);
	}
	m_Graph.AutoZoom(false, true);
}

void CDeviationsDlg::selection_changed(GtkTreeSelection *widget, CDeviationsDlg *pMe)
{
	pMe->OnSelectionChanged(widget);
}

void CDeviationsDlg::OnSelectionChanged(GtkTreeSelection *selection)
{
	int id;
	GtkTreeModel *model;
	GtkTreeIter iter;

	if (gtk_tree_selection_get_selected(selection, &model, &iter)) {
		if (model == GTK_TREE_MODEL(m_DataList)) {
			gtk_tree_model_get(model, &iter, DATA_COLUMN, &id, -1);
			SetData(id);
		}
	}
}

void CDeviationsDlg::response_dialog(GtkDialog *pDlg, gint response_id, CDeviationsDlg *pMe)
{
	if (!pMe->OnCloseQuery(response_id)) 
		g_signal_stop_emission_by_name(pDlg, "response");
}

bool CDeviationsDlg::OnCloseQuery(gint response_id)
{
	if (response_id == GTK_RESPONSE_NONE) {
		SaveData();
		return false;
	}
	return true;
}

void CDeviationsDlg::col_name_free(char *str, void *data)
{
	g_free(str);
}

void CDeviationsDlg::SaveData(void)
{
	char *filename, *folder, *fpath, *f;

	GtkWidget *pSaveDlg = gtk_file_chooser_dialog_new("Save data to...",
		GTK_WINDOW(m_pDlg), GTK_FILE_CHOOSER_ACTION_SAVE, GTK_STOCK_CANCEL, 
		GTK_RESPONSE_CANCEL, GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT, NULL);
	gtk_file_chooser_set_do_overwrite_confirmation(GTK_FILE_CHOOSER(pSaveDlg), true);

	folder = CConfig::GetStr("Output", "Folder", NULL);
	if (folder) {
		gtk_file_chooser_set_current_folder(GTK_FILE_CHOOSER(pSaveDlg), folder);
		g_free(folder);
	}
	filename = CConfig::GetStr("Deviations", "Name", "output.txt");
	if (filename) {
		gtk_file_chooser_set_current_name(GTK_FILE_CHOOSER(pSaveDlg), filename);
		g_free(filename);
	}

	if (gtk_dialog_run(GTK_DIALOG(pSaveDlg))==GTK_RESPONSE_ACCEPT) {
		folder = gtk_file_chooser_get_current_folder(GTK_FILE_CHOOSER(pSaveDlg));
		CConfig::SetStr("Output", "Folder", folder);
		g_free(folder);
		fpath = gtk_file_chooser_get_filename(GTK_FILE_CHOOSER(pSaveDlg));
		filename = g_path_get_basename(fpath);
		CConfig::SetStr("Deviations", "Name", filename);
		g_free(filename);
		f = g_locale_from_utf8(fpath, -1, NULL, NULL, NULL);
		cmpack_tab_save(m_Table, f);
		g_free(f);
		g_free(fpath);
		Global.pProject->Save();
	}
	gtk_widget_destroy(pSaveDlg);
}
