/**************************************************************

graph.cpp (C-Munipack project)
Widget which can draw a graph
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>

#include "graph_data.h"
#include "graph.h"


//
// Default constructor
//
CGraphData::CGraphData():m_Channels(0), m_RecSize(1), m_Count(0), m_Capacity(0), 
	m_Data(NULL), m_GraphCount(0), m_GraphCapacity(0), m_GraphList(NULL), 
	m_DataChanged(false), m_RedrawLockCount(0)
{
}


//
// Constructor
//
CGraphData::CGraphData(int channels, int capacity):m_Channels(channels), 
	m_RecSize(channels+1),m_Count(0), m_Capacity(capacity), m_Data(NULL), 
	m_GraphCount(0), m_GraphCapacity(0), m_GraphList(NULL), m_DataChanged(false), 
	m_RedrawLockCount(0)
{
	if (m_Capacity>0)
		m_Data = (double*)g_malloc(m_RecSize*m_Capacity*sizeof(double));
}


//
// Destructor
//
CGraphData::~CGraphData()
{
	int i;

	if (m_GraphList) {
		for (i=0; i<m_GraphCount; i++) 
			m_GraphList[i]->SetData(NULL);
		g_free(m_GraphList);
	}
	if (m_Data)
		g_free(m_Data);
}


//
// Connect graph model
//
void CGraphData::Connect(CGraph *pGraph)
{
	int i;

	for (i=0; i<m_GraphCount; i++) {
		if (m_GraphList[i] == pGraph)
			return;
	}
	if (m_GraphCount>=m_GraphCapacity) {
		m_GraphCapacity += 8;
		m_GraphList = (CGraph**)g_realloc(m_GraphList, m_GraphCapacity*sizeof(CGraph*));
	}
	m_GraphList[m_GraphCount++] = pGraph;
	pGraph->SetData(this);
}


//
// Disconnect graph model
//
void CGraphData::Disconnect(CGraph *pGraph)
{
	int i, j;

	for (i=0; i<m_GraphCount; i++) {
		if (m_GraphList[i] == pGraph) {
			for (j=i+1; j<m_GraphCount; j++) 
				m_GraphList[j-1] = m_GraphList[j];
			m_GraphCount--;
			pGraph->SetData(NULL);
			break;
		}
	}
}


//
// Clear data and set number of channels and initial capacity
//
void CGraphData::Init(int channels, int capacity)
{
	m_Channels = (channels>=0 ? channels : 0);
	m_RecSize = m_Channels + 1;
	m_Capacity = (capacity>=0 ? capacity : 0);
	m_Count = 0;
	if (m_Data) {
		if (m_Capacity==0) {
			g_free(m_Data);
			m_Data = NULL;
		} else {
			m_Data = (double*)g_realloc(m_Data, m_Capacity*m_RecSize*sizeof(double));
		}
		Invalidate();
	} else {
		if (m_Capacity>0)
			m_Data = (double*)g_malloc(m_Capacity*m_RecSize*sizeof(double));
	}
}


//
// Clear all data
//
void CGraphData::Clear()
{
	if (m_Data) {
		g_free(m_Data);
		m_Data = NULL;
	}
	m_Count = m_Capacity = 0;
	Invalidate();
}


//
// Lock redrawing
//
void CGraphData::LockRedraw()
{
	if (m_RedrawLockCount)
		m_DataChanged = false;
	m_RedrawLockCount++;
}


//
// Unlock redrawing
//
void CGraphData::UnlockRedraw()
{
	int i;

	assert(m_RedrawLockCount > 0);

	m_RedrawLockCount--;
	if (m_RedrawLockCount == 0 && m_GraphCount>0 && m_DataChanged) {
		for (i=0; i<m_GraphCount; i++)
			m_GraphList[i]->Invalidate();
		m_DataChanged = false;
	}
}


//
// Add new data to the graph
//
void CGraphData::Append(double x, const double *values)
{
	if (m_Count>=m_Capacity) {
		m_Capacity += 64;
		m_Data = (double*)g_realloc(m_Data, m_Capacity*m_RecSize*sizeof(double));
	}
	m_Data[m_Count*m_RecSize] = x;
	memcpy(m_Data+(m_Count*m_RecSize+1), values, 
		m_Channels*sizeof(double));
	m_Count++;
	Invalidate();
}


//
// Set X value
//
void CGraphData::SetX(int index, double x)
{
	if (index>=0 && index<m_Count) {
		m_Data[index*m_RecSize] = x;
		Invalidate();
	}
}


//
// Set values
//
void CGraphData::SetValues(int index, const double *values)
{
	if (index>=0 && index<m_Count) {
		memcpy(m_Data+(m_Count*m_RecSize+1), values, 
			m_Channels*sizeof(double));
		Invalidate();
	}
}


//
// Invalidate graph
//
void CGraphData::Invalidate()
{
	int i;

	if (m_RedrawLockCount==0) {
		for (i=0; i<m_GraphCount; i++)
			m_GraphList[i]->Invalidate();
	} else {
		m_DataChanged = true;
	}
}
