/**************************************************************

preview_dlg.cpp (C-Munipack project)
The preview dialog
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <math.h>

#include "phtfile_dlg.h"
#include "configuration.h"
#include "showheader_dlg.h"
#include "utils.h"

#define APERTURE_DEFAULT 0

enum tMenuId
{
	MENU_FILE = 1,
	MENU_VIEW
};

enum tCommandId
{
	CMD_CLOSE = 100,
	CMD_PRINT,
	CMD_EXPORT,
	CMD_PROPERTIES,
	CMD_REMOVE,
	CMD_BADSTARS,
	CMD_ZOOM_IN,
	CMD_ZOOM_OUT,
	CMD_ZOOM_100
};

enum tLabelId
{
	LABEL_CAPTION,
	LABEL_POSITION,
	LABEL_MAGNITUDE,
	LABEL_REFERENCE
};

enum tFieldId
{
	FIELD_FILE,
	FIELD_FILENAME,
	FIELD_DIRPATH,
	FIELD_FRAME,
	FIELD_DATETIME,
	FIELD_JULDAT,
	FIELD_FILTER,
	FIELD_EXPTIME,
	FIELD_PHOTOMETRY,
	FIELD_STARS
};

//-------------------------   MENU FOR PREVIEW WINDOW   ---------------------------

static const CMenuBar::tMenuItem FileMenu[] = {
	{ CMenuBar::MB_ITEM, CMD_PROPERTIES,			"Show _Properties" },
	{ CMenuBar::MB_SEPARATOR },
	//{ CMenuBar::MB_ITEM, CMD_EXPORT,			"E_xport" },
	//{ CMenuBar::MB_ITEM, CMD_PRINT,			"P_rint" },
	//{ CMenuBar::MB_SEPARATOR },
	{ CMenuBar::MB_ITEM, CMD_CLOSE,				"_Close" },
	{ CMenuBar::MB_END }
};

static const CMenuBar::tMenuItem ZoomMenu[] = {
	{ CMenuBar::MB_ITEM, CMD_ZOOM_IN,			"Zoom _in" },
	{ CMenuBar::MB_ITEM, CMD_ZOOM_OUT,			"Zoom _out" },
	{ CMenuBar::MB_ITEM, CMD_ZOOM_100,			"Zoom 1:1" },
	{ CMenuBar::MB_END }
};

static const CMenuBar::tMenuItem ViewMenu[] = {
	{ CMenuBar::MB_CHECKBTN, CMD_BADSTARS,		"Bad _stars" },
	{ CMenuBar::MB_SEPARATOR },
	{ CMenuBar::MB_SUBMENU, 0,					"_Zoom", NULL, ZoomMenu },
	{ CMenuBar::MB_END }
};

static const CMenuBar::tMenu PreviewMenu[] = {
	{ "_File", MENU_FILE, FileMenu },
	{ "_View", MENU_VIEW, ViewMenu },
	{ NULL }
};

//------------------------  COMBO BOX STRINGS   ---------------------------------

static GtkWidget *toolbar_new(GtkOrientation orientation)
{
	GtkWidget *tbox = gtk_toolbar_new();
	gtk_toolbar_set_style(GTK_TOOLBAR(tbox), GTK_TOOLBAR_ICONS);
	gtk_toolbar_set_orientation(GTK_TOOLBAR(tbox), orientation);
	return tbox;
}

static GtkToolItem *toolbar_new_button_from_stock(GtkWidget *toolbar, const gchar *stock_id)
{
	GtkToolItem *tbtn = gtk_tool_button_new_from_stock(stock_id);
	gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tbtn, -1);
	return tbtn;
}

static GtkToolItem *toolbar_new_widget(GtkWidget *toolbar, GtkWidget *widget)
{
	GtkToolItem *item = gtk_tool_item_new();
	gtk_container_add(GTK_CONTAINER(item), widget);
	gtk_toolbar_insert(GTK_TOOLBAR(toolbar), item, -1);
	return item;
}

static GtkWidget *toolbar_new_entry(GtkWidget *toolbar)
{
	GtkWidget *entry = gtk_entry_new();
	gtk_entry_set_width_chars(GTK_ENTRY(entry), 10);
	gtk_entry_set_alignment(GTK_ENTRY(entry), 0.5);
	gtk_editable_set_editable(GTK_EDITABLE(entry), false);
	toolbar_new_widget(toolbar, entry);
	return entry;
}

/*static GtkToolItem *toolbar_new_separator(GtkWidget *toolbar)
{
	GtkToolItem *tbtn = gtk_separator_tool_item_new();
	gtk_toolbar_insert(GTK_TOOLBAR(toolbar), tbtn, -1);
	return tbtn;
}*/

//-------------------------   PREVIEW DIALOG   --------------------------------

CPhtFileDlg::CPhtFileDlg(GtkWindow *parent):CFileDlg(parent), m_Updating(false), 
	m_LastPosX(-1), m_LastPosY(-1), m_LastIndex(-1), m_InfoMode(INFO_NONE)
{
	static const GdkGeometry geom = { 320, 240 };

	GtkWidget *tbox, *vbox;

	gtk_window_set_geometry_hints(GTK_WINDOW(m_pDlg), NULL, (GdkGeometry*)&geom, GDK_HINT_MIN_SIZE);
	gtk_window_set_default_size(GTK_WINDOW(m_pDlg), 400, 320);

	m_Aperture = CConfig::GetInt("PreviewDlg", "Aperture", APERTURE_DEFAULT);

	// Dialog icon
	gchar *icon = get_icon_file("chart");
	gtk_window_set_icon(GTK_WINDOW(m_pDlg), gdk_pixbuf_new_from_file(icon, NULL));
	g_free(icon);

	// Window layout
	vbox = gtk_vbox_new(FALSE, 0);
	gtk_container_add(GTK_CONTAINER(m_pDlg), vbox);

	// Menu bar
	m_Menu.Create(PreviewMenu, false);
	m_Menu.RegisterCallback(MenuClick, this);
	gtk_box_pack_start(GTK_BOX(vbox), m_Menu.Handle(), false, false, 0);

	// Toolbar
	tbox = toolbar_new(GTK_ORIENTATION_HORIZONTAL);
	gtk_box_pack_start(GTK_BOX(vbox), tbox, false, false, 0);
	m_ZoomOut = toolbar_new_button_from_stock(tbox, GTK_STOCK_ZOOM_OUT);
	m_ZoomEntry = toolbar_new_entry(tbox);
	m_ZoomIn = toolbar_new_button_from_stock(tbox, GTK_STOCK_ZOOM_IN);

	// Chart
	m_Chart.SetZoom(CConfig::GetDbl("PreviewDlg", "ZoomFactor", 1.0));
	m_Chart.ShowBadStars(CConfig::GetBool("PreviewDlg", "BadStars", true));
	m_Chart.ShowImage(false);
	m_Chart.ShowChart(true);
	m_Chart.EnableTracking(true);
	m_Chart.ShowNonMatchedStars(true);
	m_Chart.HighlightNonMatchedStars(false);
	m_Chart.RegisterCallback(ChartCallback, this);
	gtk_box_pack_start(GTK_BOX(vbox), GTK_WIDGET(g_object_ref(m_Chart.Handle())), true, true, 0);

	// Info box
	gtk_box_pack_start(GTK_BOX(vbox), m_InfoBox.Handle(), FALSE, FALSE, 0);
	
	// Status bar
	m_Status.Create();
	gtk_box_pack_start(GTK_BOX(vbox), m_Status.Handle(), FALSE, FALSE, 0);

	// Signals
	g_signal_connect(G_OBJECT(m_ZoomIn), "clicked", G_CALLBACK(button_clicked), this);
	g_signal_connect(G_OBJECT(m_ZoomOut), "clicked", G_CALLBACK(button_clicked), this);

	// Show the dialog
	gtk_widget_show_all(vbox);
	m_InfoBox.Show(false);
	UpdateControls();
}

bool CPhtFileDlg::LoadFile(const char *fpath)
{
	m_Chart.Clear();
	m_Chart.LoadChart(fpath, false);
	m_Chart.SetAperture(1);
	m_Status.Set(NULL);
	m_InfoBox.Show(false);
	UpdateControls();
	return true;
}

void CPhtFileDlg::UpdateControls(void)
{
	bool ok;
	char txt[256];
	
	// Frame menu
	m_Menu.EnableMenu(MENU_FILE, true);
	m_Menu.Enable(CMD_PROPERTIES, true);
	m_Menu.Enable(CMD_EXPORT, true);
	m_Menu.Enable(CMD_PRINT, true);

	// View menu
	m_Menu.EnableMenu(MENU_VIEW, true);
	ok = true;
	gtk_widget_set_sensitive(GTK_WIDGET(m_ZoomOut), ok);
	gtk_widget_set_sensitive(GTK_WIDGET(m_ZoomEntry), ok);
	gtk_widget_set_sensitive(GTK_WIDGET(m_ZoomIn), ok);
	m_Menu.Enable(CMD_ZOOM_IN, ok);
	m_Menu.Enable(CMD_ZOOM_OUT, ok);
	m_Menu.Enable(CMD_ZOOM_100, ok);
	m_Menu.Check(CMD_BADSTARS, m_Chart.BadStars());

	// Zoom factor
	sprintf(txt, "%.0f %%", 100.0*m_Chart.Zoom());
	gtk_entry_set_text(GTK_ENTRY(m_ZoomEntry), txt);
}

void CPhtFileDlg::MenuClick(int message, int wparam, void *lparam, void *cb_data)
{
	CPhtFileDlg *pMe = (CPhtFileDlg*)cb_data;

	switch (message)
	{
	case CMenuBar::CB_ACTIVATE:
		pMe->OnCommand(wparam);
		break;
	}
}

void CPhtFileDlg::OnCommand(int cmd_id)
{
	switch (cmd_id)
	{
	// File menu
	case CMD_CLOSE:
		Destroy();
		break;
	case CMD_PROPERTIES:
		ShowProperties();
		break;

	// View menu
	case CMD_BADSTARS:
		m_Chart.ShowBadStars(m_Menu.IsChecked(CMD_BADSTARS));
		CConfig::SetBool("PreviewDlg", "BadStars", m_Chart.BadStars());
		break;

	// Zoom menu
	case CMD_ZOOM_IN:
		m_Chart.ZoomIn();
		break;
	case CMD_ZOOM_OUT:
		m_Chart.ZoomOut();
		break;
	case CMD_ZOOM_100:
		m_Chart.SetZoom(1.0);
		break;
	}
}

void CPhtFileDlg::ShowProperties(void)
{
	CPhtFileInfoDlg dlg(GTK_WINDOW(m_pDlg));
	dlg.ShowModal(m_Path, m_Chart.ChartInfo());
}

void CPhtFileDlg::button_clicked(GtkWidget *button, CPhtFileDlg *pDlg)
{
	pDlg->OnButtonClicked(button);
}

void CPhtFileDlg::OnButtonClicked(GtkWidget *pBtn)
{
	if (pBtn==GTK_WIDGET(m_ZoomIn))
		m_Chart.ZoomIn();
	else if (pBtn==GTK_WIDGET(m_ZoomOut))
		m_Chart.ZoomOut();
}

void CPhtFileDlg::ChartCallback(int message, int wparam, void* lparam, void* cb_data)
{
	CPhtFileDlg *pMe = (CPhtFileDlg*)cb_data;

	switch (message)
	{
	case CChartBase::CB_ZOOM_CHANGE:
		pMe->OnZoomChanged();
		break;
	case CChartBase::CB_MOUSE_MOVE:
		pMe->OnMouseMoved();
		break;
	case CChartBase::CB_LEFT_CLICK:
		pMe->OnLButtonClick(((GdkEventButton*)lparam)->time);
		break;
	}
}


void CPhtFileDlg::OnZoomChanged(void)
{
	char buf[256];

	sprintf(buf, "%.0f %%", 100.0*m_Chart.Zoom());
	gtk_entry_set_text(GTK_ENTRY(m_ZoomEntry), buf);
}


void CPhtFileDlg::OnMouseMoved(void)
{
	char	buf[256], aux[128];
	int		x, y, index;
	CChartEx::tStarInfo star;

	if (m_Chart.GetCursorPos(&x, &y)) 
		index = m_Chart.FindStarByPos(x, y, true);
	else
		index = -1;
	if (index>=0) {
		if (m_LastIndex!=index) {
			m_LastIndex = index;
			m_Chart.SetFocusedStar(index);
			m_Chart.GetStar(index, &star);
			sprintf(buf, "Star #%d: X = %.1f, Y = %.1f", star.id, star.x, star.y);
			if (star.mag_valid && star.err_valid) {
				sprintf(aux, ", Mag = %.4f +- %.4f", star.mag, star.err);
				strcat(buf, aux);
			}
			if (star.mag_valid && !star.err_valid) {
				sprintf(aux, ", Mag = %.4f +- ????", star.mag);
				strcat(buf, aux);
			}
			if (!star.mag_valid && !star.err_valid) {
				sprintf(aux, ", Mag = ???? +- ????");
				strcat(buf, aux);
			}
			if (star.matched) {
				sprintf(aux, ", Ref. = %d", star.xref);
				strcat(buf, aux);
			}
			m_Status.Set(buf);
		}
	} else {
		if (m_LastIndex>=0) {
			m_LastIndex = -1;
			m_Chart.SetFocusedStar(-1);
			m_Status.Set(NULL);
		}
	}
}


//
// Left button click
//
void CPhtFileDlg::OnLButtonClick(guint32 timestamp)
{
	int index = m_Chart.FocusedStar();
	if (index>=0) {
		// Star info tool
		if (m_Chart.SelectedStar()!=index) {
			m_Chart.SelectStar(index);
			UpdateInfoBox();
		}
	} else {
		// Hide infobox
		if (m_Chart.SelectedStar()>=0) {
			m_Chart.SelectStar(-1);
			UpdateInfoBox();
		}
	}
}

void CPhtFileDlg::UpdateInfoBox(void)
{
	int index;
	char buf[256];
	CChartEx::tStarInfo star;

	// Show information about selected star
	index = m_Chart.SelectedStar();
	if (index>=0) {
		if (m_InfoMode != INFO_STAR) {
			m_InfoBox.Clear();
			m_InfoBox.AddHeading(LABEL_CAPTION, 0, 0);
			m_InfoBox.AddField(LABEL_POSITION, 0, 1, "Position");
			m_InfoBox.AddField(LABEL_MAGNITUDE, 0, 2, "Brightness");
			m_InfoBox.AddField(LABEL_REFERENCE, 0, 3, "Reference id");
			m_InfoMode = INFO_STAR;
		}
		m_Chart.GetStar(index, &star);
		// Caption
		switch (star.sel_type)
		{
		case ST_VARIABLE:
			sprintf(buf, "Star #%d (variable star %d)", star.id, star.sel_index);
			m_InfoBox.SetHeading(LABEL_CAPTION, buf, "red");
			break;
		case ST_COMPARISON:
			sprintf(buf, "Star #%d (comparison star %d)", star.id, star.sel_index);
			m_InfoBox.SetHeading(LABEL_CAPTION, buf, "green");
			break;
		case ST_CHECK:
			sprintf(buf, "Star #%d (check star %d)", star.id, star.sel_index);
			m_InfoBox.SetHeading(LABEL_CAPTION, buf, "blue");
			break;
		default:
			sprintf(buf, "Star #%d", star.id);
			m_InfoBox.SetHeading(LABEL_CAPTION, buf, "black");
			break;
		}
		// Position
		sprintf(buf, "%.1f, %.1f", star.x, star.y);
		m_InfoBox.SetField(LABEL_POSITION, buf);
		// Brightness
		if (star.mag_valid) {
			if (star.err_valid) 
				sprintf(buf, "%.4f +- %.4f mag", star.mag, star.err);
			else 
				sprintf(buf, "%.4f +- ???? mag", star.mag);
			m_InfoBox.SetField(LABEL_MAGNITUDE, buf);
		} else {
			m_InfoBox.SetField(LABEL_MAGNITUDE, "------");
		}
		// Reference id
		if (star.matched)
			m_InfoBox.SetField(LABEL_REFERENCE, star.xref);
		else
			m_InfoBox.SetField(LABEL_REFERENCE, "------");
		m_InfoBox.Show(true);
	} else {
		m_InfoBox.Show(false);
	}
}

//-------------------------   FILE PREVIEW DIALOG   --------------------------------

//
// Constructor
//
CPhtFileDlgModal::CPhtFileDlgModal(GtkWindow *pParent):CFileDlgModal(pParent)
{
	static const CMenuBar::tMenuItem FileMenu[] = {
		{ CMenuBar::MB_ITEM, CMD_PROPERTIES, "_Properties" },
		{ CMenuBar::MB_END }
	};

	static const CMenuBar::tMenu PreviewMenu[] = {
		{ "_File", MENU_FILE, FileMenu },
		{ NULL }
	};

	GtkWidget *vbox;

	m_Menu.Create(PreviewMenu, false);
	m_Menu.RegisterCallback(MenuClick, this);
	m_Chart.EnableTracking(true);
	m_Chart.ShowImage(false);
	m_Chart.ShowChart(true);

	// Window layout
	vbox = GTK_DIALOG(m_pDlg)->vbox;
	gtk_box_pack_start (GTK_BOX(vbox), m_Menu.Handle(), false, false, 0);
	gtk_box_pack_start(GTK_BOX(vbox), GTK_WIDGET(g_object_ref(m_Chart.Handle())), true, true, 0);
	gtk_widget_show_all(vbox);
}


//
// Menu handler
//
void CPhtFileDlgModal::MenuClick(int message, int wparam, void *lparam, void *cb_data)
{
	CPhtFileDlgModal *pMe = (CPhtFileDlgModal*)cb_data;

	switch (message)
	{
	case CMenuBar::CB_ACTIVATE:
		pMe->OnCommand(wparam);
		break;
	}
}


bool CPhtFileDlgModal::LoadFile(const char *fpath)
{
	m_Chart.Clear();
	m_Chart.LoadChart(fpath, false);
	m_Chart.SetAperture(1);
	return true;
}

void CPhtFileDlgModal::OnCommand(int cmd_id)
{
	switch (cmd_id)
	{
	case CMD_PROPERTIES:
		ShowProperties();
		break;
	}
}

void CPhtFileDlgModal::ShowProperties(void)
{
	CPhtFileInfoDlg dlg(GTK_WINDOW(m_pDlg));
	dlg.ShowModal(m_Path, m_Chart.ChartInfo());
}

//---------------------------   INFO DIALOG   --------------------------------

//
// Constructor
//
CPhtFileInfoDlg::CPhtFileInfoDlg(GtkWindow *pParent):CInfoDlg(pParent), m_Path(NULL)
{
	GtkWidget *bbox;

	AddHeading(FIELD_FILE, 0, 0, "File");
	AddField(FIELD_FILENAME, 0, 1, "File name", PANGO_ELLIPSIZE_MIDDLE);
	AddField(FIELD_DIRPATH, 0, 2, "Location", PANGO_ELLIPSIZE_MIDDLE);
	AddSeparator(0, 4);
	AddHeading(FIELD_FRAME, 0, 5, "Frame");
	AddField(FIELD_DATETIME, 0, 6, "Date and time");
	AddField(FIELD_JULDAT, 0, 7, "Julian date");
	AddField(FIELD_FILTER, 0, 8, "Optical filter", PANGO_ELLIPSIZE_END);
	AddField(FIELD_EXPTIME, 0, 9, "Exposure duration");
	AddSeparator(0, 10);
	AddHeading(FIELD_PHOTOMETRY, 0, 11, "Photometry");
	AddField(FIELD_STARS, 0, 12, "Number of stars");
	gtk_widget_show_all(m_Tab);

	bbox = gtk_hbutton_box_new();
	gtk_box_set_spacing(GTK_BOX(bbox), 8);
	gtk_container_set_border_width(GTK_CONTAINER(bbox), 4);
	m_HdrBtn = gtk_button_new_with_label("Show header...");
	gtk_box_pack_start(GTK_BOX(bbox), m_HdrBtn, true, true, 0);
	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(m_pDlg)->vbox), bbox);
	gtk_widget_show_all(bbox);

	g_signal_connect(G_OBJECT(m_HdrBtn), "clicked", G_CALLBACK(button_clicked), this);
}


//
// Destructor
//
CPhtFileInfoDlg::~CPhtFileInfoDlg()
{
	if (m_Path)
		g_free(m_Path);
}


//
// Show dialog
//
void CPhtFileInfoDlg::ShowModal(const char *path, const CChartInfo *pInfo)
{
	char buf[512], *basename, *dirpath;
	CmpackDateTime dt;

	// Update properties
	g_free(m_Path);
	m_Path = g_strdup(path);
	basename = g_path_get_basename(path);
	SetField(FIELD_FILENAME, basename);
	g_free(basename);
	dirpath = g_path_get_dirname(path);
	SetField(FIELD_DIRPATH, dirpath);
	g_free(dirpath);

	if (pInfo->JD()>0) {
		cmpack_decodejd(pInfo->JD(), &dt);
		sprintf(buf, "%04d-%02d-%02d %d:%02d:%02d.%03d", dt.date.year, dt.date.month, dt.date.day,
			dt.time.hour, dt.time.minute, dt.time.second, dt.time.milisecond);
		SetField(FIELD_DATETIME, buf);
		SetField(FIELD_JULDAT, pInfo->JD(), 7);
	} else {
		SetField(FIELD_DATETIME, "Not available");
		SetField(FIELD_JULDAT, "Not available");
	}
	if (pInfo->Filter())
		SetField(FIELD_FILTER, pInfo->Filter());
	else
		SetField(FIELD_FILTER, "");
	if (pInfo->ExpTime()>0)
		SetField(FIELD_EXPTIME, pInfo->ExpTime(), 3, "second(s)");
	else
		SetField(FIELD_EXPTIME, "Not available");

	if (pInfo->Stars()>=0)
		SetField(FIELD_STARS, pInfo->Stars());
	else
		SetField(FIELD_STARS, "Not available");

	CInfoDlg::ShowModal();
}

void CPhtFileInfoDlg::button_clicked(GtkWidget *button, CPhtFileInfoDlg *pDlg)
{
	pDlg->OnButtonClicked(button);
}

void CPhtFileInfoDlg::OnButtonClicked(GtkWidget *pBtn)
{
	if (pBtn==GTK_WIDGET(m_HdrBtn))
		ShowHeader();
}


void CPhtFileInfoDlg::ShowHeader(void)
{
	CShowHeaderDlg dlg(GTK_WINDOW(m_pDlg));
	dlg.Execute(m_Path);
}
