/**************************************************************

project.h (C-Munipack project)
Table of input frames
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#ifndef MUNIWIN_PROJECT_H
#define MUNIWIN_PROJECT_H

#include <glib.h>
#include <gtk/gtk.h>

#include "helper_classes.h"

// File state codes (bitmask)
#define	CFILE_CONVERSION	0x0001		// Fetched/converted
#define CFILE_TIMECORR		0x0002 		// Time corrected
#define CFILE_BIASCORR      0x0040		// Bias corrected
#define CFILE_DARKCORR		0x0004 		// Dark corrected
#define CFILE_FLATCORR		0x0008 		// Flat corrected
#define CFILE_PHOTOMETRY	0x0010		// Photometry processed
#define CFILE_MATCHING		0x0020		// Matching processed
#define CFILE_ERROR         0x1000		// Error

// Frame list - model columns
enum tFrameListColumns {
	FRAME_STATE,		// State bit mask
	FRAME_ID,			// File id
	FRAME_JULDAT,		// Julian date of observation
	FRAME_REPORT,		// Result of last operation
	FRAME_ORIGFILE,		// Full path of the source file 
	FRAME_TEMPFILE,		// Name of the temporary file (without path)
	FRAME_PHOTFILE,		// Name of the photometry file (without path)
	FRAME_BIASFILE,		// Full path of the bias correction frame
	FRAME_DARKFILE,		// Full path of the dark correction frame
	FRAME_FLATFILE,		// Full path of the flat correction frame
	FRAME_FILTER,		// Filter name
	FRAME_TIMECORR,		// Time correction in seconds
	FRAME_EXPTIME,		// Exposure duration in seconds
	FRAME_CCDTEMP,		// CCD temperature in centigrades
	FRAME_SUBFRAMES,	// Number of frames averaged
	FRAME_STARS,		// Number of stars found
	FRAME_MSTARS,		// Number of stars matched
	FRAME_STRINGID,		// Frame identifier as a text
	FRAME_ICON,			// Frame thumbnail 
	FRAME_NCOLS
};

//
// Project interface
//
class CProject
{
public:
	struct tStatus
	{
		int files;			// Total number of files
		int converted;		// Number of converted files
		int photometred;	// Number of photometred files
		int matched;		// Number of matched files
	};

	// Constructor
	CProject();

	// Destructor
	virtual ~CProject(void);

	// Load project from file
	void Load(void);

	// Save project to file
	void Save(void);

	// Get project status
	void GetStatus(tStatus *stat);

	// Get project's file name
	const gchar *GetFileName(void);

	// Get project's home path
	const gchar *GetDirPath(void);

	// Get list of frames
	GtkTreeModel *FileList(void)
	{ return GTK_TREE_MODEL(m_Frames); }

	// Get number of files
	int GetFileCount(void);

	// Clear project
	void Clear(void);

	// Delete all temporary files
	void ClearTempFiles(void);

	// Clear information about reference frame
	void ClearReference(void);

	// Clear information about correction frames
	void ClearCorrections(void);

	// Clear information about selected stars
	void ClearSelection(void);

	// Clear information about observed object
	void ClearObject(void);

	// Clear information about location
	void ClearObserver(void);

	// Add single file to the project
	// Returns frame id or zero on failure
	int AddFile(const gchar *filepath);

	// Returns true if the file is already in the project
	bool FileInProject(const gchar *filepath);

	// Remove single file
	void RemoveFile(GtkTreePath *pPath);

	// Remove selected files
	void RemoveFiles(GtkTreeSelection *pSel);

	// Set time correction
	void SetTimeCorrection(double seconds);

	// Set time correction
	double GetTimeCorrection(void)
	{ return m_TimeCorr; }

	// Set path to the original bias file
	void SetOrigBiasFile(const gchar *orig_file);

	// Set path to the temporary bias file
	void SetTempBiasFile(const gchar *temp_file);

	// Returns true if the temporary bias file is still valid
	bool CheckBiasFile(const gchar *orig_file);

	// Get path to the original bias file
	const gchar *GetOrigBiasFile(void)
	{ return m_OrigBiasFile.FullPath(); }

	// Get path to the temporary bias file
	const gchar *GetTempBiasFile(void)
	{ return m_TempBiasFile.FullPath(); }

	// Set path to the original dark file
	void SetOrigDarkFile(const gchar *orig_file);

	// Set path to the temporary dark file
	void SetTempDarkFile(const gchar *temp_file);

	// Returns true if the temporary dark file is still valid
	bool CheckDarkFile(const gchar *orig_file);

	// Get path to the original dark file
	const gchar *GetOrigDarkFile(void)
	{ return m_OrigDarkFile.FullPath(); }

	// Get path to the temporary dark file
	const gchar *GetTempDarkFile(void)
	{ return m_TempDarkFile.FullPath(); }
	
	// Set path to the original flat file
	void SetOrigFlatFile(const gchar *orig_file);

	// Set path to the temporary flat file
	void SetTempFlatFile(const gchar *temp_file);

	// Returns true if the temporary dark file is still valid
	bool CheckFlatFile(const gchar *orig_file);

	// Get path to the original flat file
	const gchar *GetOrigFlatFile(void)
	{ return m_OrigFlatFile.FullPath(); }

	// Get path to the temporary flat file
	const gchar *GetTempFlatFile(void)
	{ return m_TempFlatFile.FullPath(); }

	// Set reference 
	// To set reference to a frame, use frame_id > 0 and cat_file == NULL
	// To set reference to a catalog file, use frame = 0 and cat_file != NULL
	// To clear the reference, use frame = 0 and cat_file == NULL
	void SetReference(int frame_id, const gchar *cat_file);

	// Get reference 
	// If reference is a frame, frame_id is set to value > 0, otherwise to value = 0.
	// If reference is a catalog file, cat_file is set to original file path, otherwise to NULL
	// Returns true if the reference has been set.
	bool GetReference(int *frame_id, const gchar **cat_file);

	// Set path to the temporary catalogue file
	void SetTempCatFile(const gchar *temp_file);

	// Returns true if the temporary catalogue file is still valid
	bool CheckCatFile(const gchar *orig_file);

	// Get path to the temporary catalogue file
	const gchar *GetTempCatFile(void)
	{ return m_TempCatFile.FullPath(); }

	// Get tree path to the file
	GtkTreePath *GetFilePath(int frame_id);

	// Get frame id from the tree path
	int GetFrameID(GtkTreePath *pPath);
	int GetFrameID(GtkTreeIter *pIter);

	// Get state
	int GetState(int frame_id);
	int GetState(GtkTreePath *pPath);
	int GetState(GtkTreeIter *pIter);

	// Get source CCD image (full path)
	gchar *GetSourceFile(int frame_id);
	gchar *GetSourceFile(GtkTreePath *pPath);
	gchar *GetSourceFile(GtkTreeIter *pIter);

	// Get image file (full path)
	gchar *GetImageFile(int frame_id);
	gchar *GetImageFile(GtkTreePath *pPath);
	gchar *GetImageFile(GtkTreeIter *pIter);

	// Get image file base name (without path)
	gchar *GetImageFileName(int frame_id);
	gchar *GetImageFileName(GtkTreePath *pPath);
	gchar *GetImageFileName(GtkTreeIter *pIter);

	// Get photometry file (full path)
	gchar *GetPhotFile(int frame_id);
	gchar *GetPhotFile(GtkTreePath *pPath);
	gchar *GetPhotFile(GtkTreeIter *pIter);

	// Get photometry file base name (without path)
	gchar *GetPhotFileName(int frame_id);
	gchar *GetPhotFileName(GtkTreePath *pPath);
	gchar *GetPhotFileName(GtkTreeIter *pIter);

	// Set error flag and set the report string
	void SetError(int frame_id, const gchar *message, 
		int state_mask = 0, int state_value = 0);
	void SetError(GtkTreePath *pPath, const gchar *message, 
		int state_mask = 0, int state_value = 0);
	void SetError(GtkTreeIter *pIter, const gchar *message, 
		int state_mask = 0, int state_value = 0);

	// Set error flag and set the report string
	void SetResult(int frame_id, const gchar *message, 
		int state_mask = 0, int state_value = 0);
	void SetResult(GtkTreePath *pPath, const gchar *message, 
		int state_mask = 0, int state_value = 0);
	void SetResult(GtkTreeIter *pIter, const gchar *message, 
		int state_mask = 0, int state_value = 0);

	// Get tree path to the previous file
	GtkTreePath *GetFirstFile(void);

	// Get tree path to the last file
	GtkTreePath *GetLastFile(void);

	// Get tree path to the next file
	GtkTreePath *GetNextFile(GtkTreePath *path);

	// Get tree path to the previous file
	GtkTreePath *GetPreviousFile(GtkTreePath *path);

	// Update image thumbnail
	void UpdateIcon(GtkTreePath *pPath);

	// Set list of selected stars
	void SetSelection(const CSelection *sel);
	void SetCatFileSelection(const CSelection *sel);
	void SetVarFindSelection(const CSelection *sel);

	// Get list of selected stars
	CSelection *Selection(void)
	{ return &m_Selection; }
	CSelection *CatFileSelection(void)
	{ return &m_CatFileSelection; }
	CSelection *VarFindSelection(void)
	{ return &m_VarFindSelection; }

	// Set active aperture
	void SetAperture(int aper_id)
	{ m_Aperture = aper_id; }

	// Get active aperture
	int Aperture(void)
	{ return m_Aperture; }

	// Set object's coordinates
	void SetObjectCoords(const CObjectCoords *obj);
	void SetReferenceObjectCoords(const CObjectCoords *obj);

	// Get object's coordinates
	CObjectCoords *ObjectCoords(void)
	{ return &m_ObjectCoords; }
	CObjectCoords *ReferenceObjectCoords(void)
	{ return &m_RefObjectCoords; }

	// Set observer's coordinates
	void SetUserLocation(const CObserverCoords *obj);
	void SetReferenceLocation(const CObserverCoords *obj);
	void SetUseDefaultLocation(bool def);

	// Get observer's coordinates
	CObserverCoords *UserLocation(void)
	{ return &m_UserLocation; }
	CObserverCoords *ReferenceLocation(void)
	{ return &m_RefLocation; }
	bool UseDefaultLocation(void) const
	{ return m_DefaultLocation; }

	CObserverCoords DefaultLocation(void) const;

	// Configuration has been changed, reload values
	void OnConfigChange(void);

	// Update image thumbnail
	void UpdateAllIcons(void);

private:
	GtkListStore	*m_Frames;			// List of input files
	int				m_FreeID;			// Next free id
	int				m_RefID;			// Reference frame id (0 = catalogue file)
	int				m_Aperture;			// Active aperture id
	double			m_TimeCorr;			// Last time correction (seconds)
	CFileInfo		m_OrigBiasFile;		// Original bias file
	CFileInfo		m_TempBiasFile;		// Temporary bias file
	CFileInfo		m_OrigDarkFile;		// Original dark file
	CFileInfo		m_TempDarkFile;		// Temporary dark file
	CFileInfo		m_OrigFlatFile;		// Original flat file
	CFileInfo		m_TempFlatFile;		// Temporary flat file
	CFileInfo		m_OrigCatFile;		// Original catalogue file
	CFileInfo		m_TempCatFile;		// Temporary catalogue file
	CSelection		m_Selection;		// Last result of the 'Choose stars' tool
	CSelection		m_CatFileSelection;	// Selection from the catalogue file
	CSelection		m_VarFindSelection;	// Last result of 'Find variables' tool
	CObjectCoords	m_ObjectCoords;		// Object's coordinates
	CObjectCoords	m_RefObjectCoords;	// Reference object's coordinates
	bool			m_DefaultLocation;	// Use default location
	CObserverCoords	m_UserLocation;		// Observer's coordinates
	CObserverCoords	m_RefLocation;		// Reference observer's coordinates

	static GdkPixbuf *MakeIcon(const gchar *filename);
	static bool MakeFileInfo(const gchar *filename, GTimeVal *time);
	static gboolean foreach_find_file(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data);
	static gboolean foreach_update_icon(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, gpointer data);
	static gboolean make_status(GtkTreeModel *model, GtkTreePath *path, GtkTreeIter *iter, tStatus *st);
};

#endif
