/**************************************************************

preview_dlg.cpp (C-Munipack project)
The preview dialog
Copyright (C) 2008 David Motl, dmotl@volny.cz

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

**************************************************************/

#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <cmunipack.h>

#include "showheader_dlg.h"
#include "utils.h"

const static struct {
	const char *caption;		// Column name
} TreeViewColumns[] = {
	{ "Name" },
	{ "Value" },
	{ "Comment" },
	{ NULL }
};

//-------------------------   CHOOSE STARS DIALOG   --------------------------------

CShowHeaderDlg::CShowHeaderDlg(GtkWindow *pParent)
{
	int i;

	// Dialog with buttons
	m_Dlg = gtk_dialog_new_with_buttons("", pParent, 
		(GtkDialogFlags)(GTK_DIALOG_MODAL | GTK_DIALOG_DESTROY_WITH_PARENT | GTK_DIALOG_NO_SEPARATOR),
		GTK_STOCK_CLOSE, GTK_RESPONSE_CLOSE, NULL);

	// Make tree store
	m_Data = gtk_tree_store_new(3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);

	// Make tree view
	m_View = gtk_tree_view_new();
	for (i=0; TreeViewColumns[i].caption!=NULL; i++) {
		GtkTreeViewColumn *col = gtk_tree_view_column_new();
		// Set column name and alignment
		gtk_tree_view_column_set_title(col, TreeViewColumns[i].caption);
		gtk_tree_view_append_column(GTK_TREE_VIEW(m_View), col);
		// Add text renderer
		GtkCellRenderer *renderer = gtk_cell_renderer_text_new();
		gtk_tree_view_column_pack_start(col, renderer, TRUE);
		gtk_tree_view_column_add_attribute(col, renderer, "text", i);
		gtk_tree_view_set_model(GTK_TREE_VIEW(m_View), GTK_TREE_MODEL(m_Data));
	}

	// Scrolled window
	m_ScrWnd = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(m_ScrWnd),
		GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(m_ScrWnd), 
		GTK_SHADOW_ETCHED_IN);
	gtk_container_add(GTK_CONTAINER(m_ScrWnd), m_View);
	gtk_widget_set_size_request(m_ScrWnd, 480, 320);

	gtk_container_add(GTK_CONTAINER(GTK_DIALOG(m_Dlg)->vbox), m_ScrWnd);
	gtk_widget_show_all(m_ScrWnd);
}

CShowHeaderDlg::~CShowHeaderDlg()
{
	gtk_widget_destroy(m_Dlg);
	g_object_unref(m_Data);
}

void CShowHeaderDlg::Execute(const char *filename)
{
	bool ok = false;
	char buf[512];

	// Start update of multiple rows
	GtkTreeModel *model = gtk_tree_view_get_model(GTK_TREE_VIEW(m_View));
	g_object_ref(model);
	gtk_tree_view_set_model(GTK_TREE_VIEW(m_View), NULL);
	gtk_tree_store_clear(m_Data);
	switch (FileType(filename))
	{
	case FILE_TYPE_IMAGE:
		LoadImageHeader(filename);
		ok = true;
		break;
	case FILE_TYPE_PHOT:
		LoadPhotHeader(filename);
		ok = true;
		break;
	case FILE_TYPE_CAT:
		LoadCatalogHeader(filename);
		ok = true;
		break;
	case FILE_TYPE_TABLE:
		LoadTableHeader(filename);
		ok = true;
		break;
	case FILE_TYPE_READALL:
		LoadReadallHeader(filename);
		ok = true;
		break;
	default:
		ok = false;
	}
	// Finish update of multiple files
	gtk_tree_view_set_model(GTK_TREE_VIEW(m_View), model);
	g_object_unref(model);
	gtk_tree_view_expand_all(GTK_TREE_VIEW(m_View));
	
	if (ok) {
		// Set window caption
		gchar *fname = g_path_get_basename(filename);
		if (fname) {
			sprintf(buf, "%s - header", fname);
			g_free(fname);
		} else {
			strcpy(buf, "Header");
		}
		gtk_window_set_title(GTK_WINDOW(m_Dlg), buf);
		// Show the dialog
		gtk_dialog_run(GTK_DIALOG(m_Dlg));
	}
}

bool CShowHeaderDlg::LoadImageHeader(const char *filename)
{
	bool	res = false;
	int		i;
	char	*key, *val, *com;
	CmpackCcdFile *file;
	GtkTreeIter iter;

	char *f = g_locale_from_utf8(filename, -1, NULL, NULL, NULL);
	if (cmpack_ccd_open(&file, f, CMPACK_OPEN_HEADONLY)==0) {
		for (i=0; cmpack_ccd_get_param(file, i, &key, &val, &com)==0; i++) {
			if ((key && strlen(key)>0) || (val && strlen(val)>0) || (com && strlen(com)>0)) {
				gtk_tree_store_append(m_Data, &iter, NULL);
				if (key) {
					gchar *aux = g_locale_to_utf8(key, -1, NULL, NULL, NULL);
					gtk_tree_store_set(m_Data, &iter, 0, aux, -1);
					g_free(aux);
				}
				if (val) {
					gchar *aux = g_locale_to_utf8(val, -1, NULL, NULL, NULL);
					gtk_tree_store_set(m_Data, &iter, 1, aux, -1);
					g_free(aux);
				}
				if (com) {
					gchar *aux = g_locale_to_utf8(com, -1, NULL, NULL, NULL);
					gtk_tree_store_set(m_Data, &iter, 2, aux, -1);
					g_free(aux);
				}
			}
			cmpack_free(key);
			cmpack_free(val);
			cmpack_free(com);
		}
		cmpack_unref(file);
		res = true;
	}
	g_free(f);
	return res;
}

bool CShowHeaderDlg::LoadPhotHeader(const char *filename)
{
	bool	res = false;
	int		i;
	const char *key, *val, *com;
	GtkTreeIter iter;
	CmpackPhtFile *file;

	// Fill the table
	char *f = g_locale_from_utf8(filename, -1, NULL, NULL, NULL);
	if (cmpack_pht_load(&file, f, CMPACK_LOAD_HEADONLY)==0) {
		for (i=0; cmpack_pht_gkyn(file, i, &key, &val, &com); i++) {
			if ((key && strlen(key)>0) || (val && strlen(val)>0) || (com && strlen(com)>0)) {
				gtk_tree_store_append(m_Data, &iter, NULL);
				if (key) {
					gchar *aux = g_locale_to_utf8(key, -1, NULL, NULL, NULL);
					if (aux) {
						gtk_tree_store_set(m_Data, &iter, 0, aux, -1);
						g_free(aux);
					}
				}
				if (val) {
					gchar *aux = g_locale_to_utf8(val, -1, NULL, NULL, NULL);
					if (aux) {
						gtk_tree_store_set(m_Data, &iter, 1, aux, -1);
						g_free(aux);
					}
				}
				if (com) {
					gchar *aux = g_locale_to_utf8(com, -1, NULL, NULL, NULL);
					if (aux) {
						gtk_tree_store_set(m_Data, &iter, 2, aux, -1);
						g_free(aux);
					}
				}
			}
		}
		cmpack_unref(file);
		res = true;
	}
	g_free(f);
	return res;
}

bool CShowHeaderDlg::LoadCatalogHeader(const char *filename)
{
	bool	res = false;
	int		i;
	const char *key, *val, *com;
	GtkTreeIter iter;
	CmpackCatFile *file;

	// Fill the table
	char *f = g_locale_from_utf8(filename, -1, NULL, NULL, NULL);
	if (cmpack_cat_load(&file, f, CMPACK_LOAD_HEADONLY)==0) {
		for (i=0; cmpack_cat_gkyn(file, i, &key, &val, &com); i++) {
			if ((key && strlen(key)>0) || (val && strlen(val)>0) || (com && strlen(com)>0)) {
				gtk_tree_store_append(m_Data, &iter, NULL);
				if (key) {
					gchar *aux = g_locale_to_utf8(key, -1, NULL, NULL, NULL);
					if (aux) {
						gtk_tree_store_set(m_Data, &iter, 0, aux, -1);
						g_free(aux);
					}
				}
				if (val) {
					gchar *aux = g_locale_to_utf8(val, -1, NULL, NULL, NULL);
					if (aux) {
						gtk_tree_store_set(m_Data, &iter, 1, aux, -1);
						g_free(aux);
					}
				}
				if (com) {
					gchar *aux = g_locale_to_utf8(com, -1, NULL, NULL, NULL);
					if (aux) {
						gtk_tree_store_set(m_Data, &iter, 2, aux, -1);
						g_free(aux);
					}
				}
			}
		}
		cmpack_unref(file);
		res = true;
	}
	g_free(f);
	return res;
}

bool CShowHeaderDlg::LoadTableHeader(const char *filename)
{
	bool	res = false;
	int		i;
	const char *key, *val;
	CmpackTable *table;
	GtkTreeIter iter;

	// Fill the table
	char *f = g_locale_from_utf8(filename, -1, NULL, NULL, NULL);
	if (cmpack_tab_load(&table, f, CMPACK_LOAD_HEADONLY)==0) {
		for (i=0; cmpack_tab_gkyn(table, i, &key, &val); i++) {
			if ((key && strlen(key)>0) || (val && strlen(val)>0)) {
				gtk_tree_store_append(m_Data, &iter, NULL);
				if (key) {
					gchar *aux = g_locale_to_utf8(key, -1, NULL, NULL, NULL);
					if (aux) {
						gtk_tree_store_set(m_Data, &iter, 0, aux, -1);
						g_free(aux);
					}
				}
				if (val) {
					gchar *aux = g_locale_to_utf8(val, -1, NULL, NULL, NULL);
					if (aux) {
						gtk_tree_store_set(m_Data, &iter, 1, aux, -1);
						g_free(aux);
					}
				}
			}
		}
		cmpack_unref(table);
		res = true;
	}
	g_free(f);
	return res;
}

bool CShowHeaderDlg::LoadReadallHeader(const char *filename)
{
	bool	res = false;
	int		aperture, helcor;
	const char *filter;
	char	buf[256];
	CmpackAllFile *file;
	GtkTreeIter iter;

	// Fill the table
	char *f = g_locale_from_utf8(filename, -1, NULL, NULL, NULL);
	if (cmpack_all_load(&file, f, CMPACK_LOAD_HEADONLY)==0) {
		aperture = cmpack_all_get_aperture(file);
		if (aperture) {
			sprintf(buf, "%d", aperture);
			gtk_tree_store_append(m_Data, &iter, NULL);
			gtk_tree_store_set(m_Data, &iter, 0, "Aperture", 1, buf, -1);
		}
		filter = cmpack_all_get_filter(file);
		if (filter) {
			gchar *aux = g_locale_to_utf8(filter, -1, NULL, NULL, NULL);
			gtk_tree_store_append(m_Data, &iter, NULL);
			gtk_tree_store_set(m_Data, &iter, 0, "Filter", 1, aux, -1);
			g_free(aux);
		}
		helcor = cmpack_all_get_helcor(file);
		if (helcor) {
			gtk_tree_store_append(m_Data, &iter, NULL);
			gtk_tree_store_set(m_Data, &iter, 0, "Heliocentric correction", 1, "Yes", -1);
		}
		cmpack_unref(file);
		res = true;
	}
	g_free(f);
	return res;
}
